# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/validation/validator__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import entity__client_pb2 as tecton__proto_dot_args_dot_entity____client__pb2
from tecton_proto.args import feature_service__client_pb2 as tecton__proto_dot_args_dot_feature__service____client__pb2
from tecton_proto.args import feature_view__client_pb2 as tecton__proto_dot_args_dot_feature__view____client__pb2
from tecton_proto.args import transformation__client_pb2 as tecton__proto_dot_args_dot_transformation____client__pb2
from tecton_proto.args import virtual_data_source__client_pb2 as tecton__proto_dot_args_dot_virtual__data__source____client__pb2
from tecton_proto.common import schema__client_pb2 as tecton__proto_dot_common_dot_schema____client__pb2
from tecton_proto.common import spark_schema__client_pb2 as tecton__proto_dot_common_dot_spark__schema____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n/tecton_proto/validation/validator__client.proto\x12\x17tecton_proto.validation\x1a&tecton_proto/args/entity__client.proto\x1a/tecton_proto/args/feature_service__client.proto\x1a,tecton_proto/args/feature_view__client.proto\x1a.tecton_proto/args/transformation__client.proto\x1a\x33tecton_proto/args/virtual_data_source__client.proto\x1a(tecton_proto/common/schema__client.proto\x1a.tecton_proto/common/spark_schema__client.proto\"I\n\x14\x45ntityValidationArgs\x12\x31\n\x04\x61rgs\x18\x01 \x01(\x0b\x32\x1d.tecton_proto.args.EntityArgsR\x04\x61rgs\"\xeb\x01\n\x1fVirtualDataSourceValidationArgs\x12<\n\x04\x61rgs\x18\x01 \x01(\x0b\x32(.tecton_proto.args.VirtualDataSourceArgsR\x04\x61rgs\x12\x43\n\x0c\x62\x61tch_schema\x18\x02 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0b\x62\x61tchSchema\x12\x45\n\rstream_schema\x18\x03 \x01(\x0b\x32 .tecton_proto.common.SparkSchemaR\x0cstreamSchema\"Y\n\x1cTransformationValidationArgs\x12\x39\n\x04\x61rgs\x18\x01 \x01(\x0b\x32%.tecton_proto.args.TransformationArgsR\x04\x61rgs\"\xe5\x01\n\x19\x46\x65\x61tureViewValidationArgs\x12\x36\n\x04\x61rgs\x18\x01 \x01(\x0b\x32\".tecton_proto.args.FeatureViewArgsR\x04\x61rgs\x12<\n\x0bview_schema\x18\x02 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\nviewSchema\x12R\n\x16materialization_schema\x18\x03 \x01(\x0b\x32\x1b.tecton_proto.common.SchemaR\x15materializationSchema\"Y\n\x1c\x46\x65\x61tureServiceValidationArgs\x12\x39\n\x04\x61rgs\x18\x01 \x01(\x0b\x32%.tecton_proto.args.FeatureServiceArgsR\x04\x61rgs\"\xeb\x03\n\x11\x46\x63oValidationArgs\x12j\n\x13virtual_data_source\x18\x01 \x01(\x0b\x32\x38.tecton_proto.validation.VirtualDataSourceValidationArgsH\x00R\x11virtualDataSource\x12G\n\x06\x65ntity\x18\x02 \x01(\x0b\x32-.tecton_proto.validation.EntityValidationArgsH\x00R\x06\x65ntity\x12W\n\x0c\x66\x65\x61ture_view\x18\x03 \x01(\x0b\x32\x32.tecton_proto.validation.FeatureViewValidationArgsH\x00R\x0b\x66\x65\x61tureView\x12`\n\x0f\x66\x65\x61ture_service\x18\x04 \x01(\x0b\x32\x35.tecton_proto.validation.FeatureServiceValidationArgsH\x00R\x0e\x66\x65\x61tureService\x12_\n\x0etransformation\x18\x05 \x01(\x0b\x32\x35.tecton_proto.validation.TransformationValidationArgsH\x00R\x0etransformationB\x05\n\x03\x66\x63o\"h\n\x11ValidationRequest\x12S\n\x0fvalidation_args\x18\x01 \x03(\x0b\x32*.tecton_proto.validation.FcoValidationArgsR\x0evalidationArgsB\x19\n\x15\x63om.tecton.validationP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.validation.validator__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\025com.tecton.validationP\001'
  _ENTITYVALIDATIONARGS._serialized_start=402
  _ENTITYVALIDATIONARGS._serialized_end=475
  _VIRTUALDATASOURCEVALIDATIONARGS._serialized_start=478
  _VIRTUALDATASOURCEVALIDATIONARGS._serialized_end=713
  _TRANSFORMATIONVALIDATIONARGS._serialized_start=715
  _TRANSFORMATIONVALIDATIONARGS._serialized_end=804
  _FEATUREVIEWVALIDATIONARGS._serialized_start=807
  _FEATUREVIEWVALIDATIONARGS._serialized_end=1036
  _FEATURESERVICEVALIDATIONARGS._serialized_start=1038
  _FEATURESERVICEVALIDATIONARGS._serialized_end=1127
  _FCOVALIDATIONARGS._serialized_start=1130
  _FCOVALIDATIONARGS._serialized_end=1621
  _VALIDATIONREQUEST._serialized_start=1623
  _VALIDATIONREQUEST._serialized_end=1727
# @@protoc_insertion_point(module_scope)



import datetime
import logging
from typing import Optional

import attrs
import pendulum
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql import functions

from tecton_core.compute_mode import ComputeMode
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query import builder
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query_consts import anchor_time
from tecton_core.time_utils import convert_timedelta_for_version
from tecton_core.time_utils import convert_timestamp_for_version
from tecton_spark.data_observability import MetricsCollector
from tecton_spark.offline_store import get_offline_store_reader
from tecton_spark.query import translate


MATERIALIZED_RAW_DATA_END_TIME = "_materialized_raw_data_end_time"
logger = logging.getLogger(__name__)


@attrs.define(auto_attribs=True)
class MaterializationPlan:
    """Computes dataframes required for materialization to offline and online stores.

    This class contains `offline_store_data_frame` and `online_store_data_frame`, but both might not always be necessary.

    Both `offline_store_data_frame` and `online_store_data_frame` should be used during batch materialization of BFVs and BWAFVs.
    Only `offline_store_data_frame` should be used during batch materialization of SFVs and SWAFVs.
    Only `online_store_data_frame` should be used during stream materialization of SFVs and SWAFVs.
    """

    _fd: FeatureDefinitionWrapper
    base_data_frame: DataFrame
    count: Optional[int] = None

    @classmethod
    def from_querytree(
        cls, fd: FeatureDefinitionWrapper, query_tree: NodeRef, spark: SparkSession
    ) -> "MaterializationPlan":
        logger.info("Query Plan:")
        logger.info(query_tree.pretty_str())
        # TODO(sanika): Consider adding validation that the expected materialization schema matches the actual materialization schema
        spark_df = translate.spark_convert(query_tree, spark).to_dataframe(spark)
        return cls(fd=fd, base_data_frame=spark_df)

    @classmethod
    def from_parquet(cls, fd: FeatureDefinitionWrapper, path: str, spark: SparkSession) -> "MaterializationPlan":
        return cls(fd=fd, base_data_frame=spark.read.parquet(path))

    @classmethod
    def from_offline_store(
        cls,
        fd: FeatureDefinitionWrapper,
        start_time: datetime.datetime,
        end_time: datetime.datetime,
        spark: SparkSession,
    ) -> Optional["MaterializationPlan"]:
        from pyspark.sql.utils import AnalysisException

        try:
            version = fd.get_feature_store_format_version
            offline_store_reader = get_offline_store_reader(spark, fd)
            offline_store_df = offline_store_reader.read(
                # subtract 1us from end time because offline store reader is inclusive on end time
                pendulum.period(start_time, end_time - datetime.timedelta(microseconds=1)),
            )
            start_time_col = convert_timestamp_for_version(start_time, version)
            end_time_col = convert_timestamp_for_version(end_time, version)
            offline_store_df = offline_store_df.filter(
                (functions.col(anchor_time()) >= functions.lit(start_time_col))
                & (functions.col(anchor_time()) < functions.lit(end_time_col))
            )
            return cls(fd=fd, base_data_frame=offline_store_df)
        except AnalysisException as e:
            if "Unable to infer schema for Parquet. It must be specified manually." in str(e):
                logger.warning("Unable to infer Parquet schema; assuming offline store is empty")
                return None
            else:
                raise e

    @property
    def offline_store_data_frame(self) -> DataFrame:
        return self.base_data_frame

    @property
    def online_store_data_frame(self) -> DataFrame:
        online_df = self.base_data_frame

        # batch online and offline df are slightly different
        if not self._fd.compaction_enabled and self._fd.is_temporal and not online_df.isStreaming:
            version = self._fd.get_feature_store_format_version
            batch_mat_schedule = convert_timedelta_for_version(self._fd.batch_materialization_schedule, version)
            online_df = self.base_data_frame.withColumn(
                MATERIALIZED_RAW_DATA_END_TIME, functions.col(anchor_time()) + batch_mat_schedule
            ).drop(anchor_time())
        return online_df

    def cached_count(self):
        if self.count is None:
            self.count = self.base_data_frame.count()
        return self.count


def get_batch_materialization_plan(
    *,
    spark: SparkSession,
    feature_definition: FeatureDefinitionWrapper,
    feature_data_time_limits: Optional[pendulum.Period],
    metrics_collector: Optional[MetricsCollector] = None,
) -> MaterializationPlan:
    """
    NOTE: We rely on Spark's lazy evaluation model to infer partially materialized tile Schema during FeatureView
    creation time without actually performing any materialization.
    Please make sure to not perform any Spark operations under this function's code path that will actually execute
    the Spark query (e.g: df.count(), df.show(), etc.).
    """
    query_tree = builder.build_materialization_querytree(
        Dialect.SPARK,
        ComputeMode.SPARK,
        feature_definition,
        for_stream=False,
        feature_data_time_limits=feature_data_time_limits,
        enable_feature_metrics=(metrics_collector is not None),
    )
    return MaterializationPlan.from_querytree(fd=feature_definition, query_tree=query_tree, spark=spark)


def get_batch_compaction_online_materialization_plan(
    *, spark: SparkSession, feature_definition: FeatureDefinitionWrapper, compaction_job_end_time: datetime.datetime
) -> MaterializationPlan:
    """The materialization plan for compaction jobs to the online store. Only used by feature views with compaction enabled.

    NOTE: We rely on Spark's lazy evaluation model to infer partially materialized tile Schema during FeatureView
    creation time without actually performing any materialization.
    Please make sure to not perform any Spark operations under this function's code path that will actually execute
    the Spark query (e.g: df.count(), df.show(), etc.).
    """
    query_tree = builder.build_compaction_query(
        Dialect.SPARK, ComputeMode.SPARK, feature_definition, compaction_job_end_time
    )
    return MaterializationPlan.from_querytree(fd=feature_definition, query_tree=query_tree, spark=spark)


def get_stream_materialization_plan(
    *,
    spark: SparkSession,
    feature_definition: FeatureDefinitionWrapper,
    metrics_collector: Optional[MetricsCollector] = None,
) -> MaterializationPlan:
    query_tree = builder.build_materialization_querytree(
        Dialect.SPARK,
        ComputeMode.SPARK,
        feature_definition,
        for_stream=True,
        enable_feature_metrics=(metrics_collector is not None),
    )
    return MaterializationPlan.from_querytree(fd=feature_definition, query_tree=query_tree, spark=spark)



import importlib
import os


def __init():
    # Workaround for https://github.com/sdispater/pendulum/issues/429
    # to allows loading this module from a .zip with transitive dependencies included. It works by tricking an
    # additional check that pendulum does using os.path.exists() before it imports the pendulum.locales.en.locale module
    real_os_path_exists = os.path.exists
    try:

        def fake_os_path_exists(path, *args, **kwargs):
            return path.endswith(".zip/pendulum/locales/en") or real_os_path_exists(path, *args, **kwargs)

        os.path.exists = fake_os_path_exists
        importlib.import_module("pendulum")
    finally:
        os.path.exists = real_os_path_exists


__init()



import py4j

from tecton_core.errors import AccessError


def handleDataAccessErrors(func, details):
    try:
        return func()
    except Exception as e:
        if isinstance(
            e, py4j.protocol.Py4JJavaError
        ) and "com.amazonaws.services.s3.model.AmazonS3Exception: Forbidden" in str(e):
            msg = f"Unable to access file: {details}"
            raise AccessError(msg) from e
        raise e



from typing import Any
from typing import Dict

import pandas
import pyspark
import pyspark.sql.types


def pandas_to_spark(
    spark: pyspark.sql.session.SparkSession, pandas_df: pandas.DataFrame, **kwargs: Dict[str, Any]
) -> pyspark.sql.DataFrame:
    # Prior to Spark 3.4, pyspark attempts to call iteritems when converting a pandas DF to a Spark DF. In Pandas 2.0,
    # iteritems() has been deleted, but items() does the same thing. We can patch over this incompatibility by copying
    # the pandas DF and aliasing iteritems to items.
    #
    # We can't easily check pyspark.__version__ directly, because databricks-connect installs a version of the pyspark
    # module which uses a different versioning scheme. Instead, we check for the existence of a symbol which was first
    # introduced in pyspark 3.4
    # TODO(liangqi): Check pyspark.__version__ directly when we migrate off of databricks-connect.
    is_old_pyspark = not hasattr(pyspark.sql.types, "JVMView")
    if not hasattr(pandas_df, "iteritems") and is_old_pyspark:
        pandas_df = pandas_df.copy(deep=False)
        pandas_df.iteritems = pandas_df.items
    return spark.createDataFrame(pandas_df, **kwargs)



import urllib
from typing import Optional

import ray
from google.protobuf import timestamp_pb2

from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.ray.delta import DeltaWriter
from tecton_materialization.ray.delta import OfflineStoreParams
from tecton_proto.data.feature_view__client_pb2 import FeatureView
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.online_store_writer.config__client_pb2 import OnlineStoreWriterConfiguration
from tecton_proto.online_store_writer.copier__client_pb2 import GCSStage
from tecton_proto.online_store_writer.copier__client_pb2 import LocalFileStage
from tecton_proto.online_store_writer.copier__client_pb2 import ObjectCopyRequest
from tecton_proto.online_store_writer.copier__client_pb2 import OnlineStoreCopierRequest
from tecton_proto.online_store_writer.copier__client_pb2 import S3Stage
from tecton_proto.online_store_writer.copier__client_pb2 import StatusUpdateRequest
from tecton_proto.online_store_writer.copier__client_pb2 import TimestampUnit


def run_online_store_copier(request):
    request_bytes = request.SerializeToString()
    runner_function = ray.cross_language.java_function(
        "com.tecton.onlinestorewriter.OnlineStoreCopier", "runFromSerializedRequest"
    )
    job = runner_function.remote(request_bytes, None)
    ray.get(job)


def write_to_online_store(
    online_store_writer_config: OnlineStoreWriterConfiguration,
    feature_view: FeatureView,
    feature_end_timestamp: timestamp_pb2.Timestamp,
    fd: FeatureDefinitionWrapper,
    stage_uri_str: str,
) -> None:
    stage_uri = urllib.parse.urlparse(stage_uri_str)
    if stage_uri.scheme in ("file", ""):
        request = OnlineStoreCopierRequest(
            online_store_writer_configuration=online_store_writer_config,
            feature_view=feature_view,
            object_copy_request=ObjectCopyRequest(
                local_file_stage=LocalFileStage(location=stage_uri.path), timestamp_units=TimestampUnit.MICROS
            ),
        )
    elif stage_uri.scheme == "s3":
        key = stage_uri.path
        if key.startswith("/"):
            key = key[1:]
        request = OnlineStoreCopierRequest(
            online_store_writer_configuration=online_store_writer_config,
            feature_view=feature_view,
            object_copy_request=ObjectCopyRequest(
                s3_stage=S3Stage(
                    bucket=stage_uri.netloc,
                    key=key,
                ),
                timestamp_units=TimestampUnit.MICROS,
            ),
        )
    elif stage_uri.scheme == "gs":
        blob = stage_uri.path
        if blob.startswith("/"):
            blob = blob[1:]
        request = OnlineStoreCopierRequest(
            online_store_writer_configuration=online_store_writer_config,
            feature_view=feature_view,
            object_copy_request=ObjectCopyRequest(
                gcs_stage=GCSStage(
                    bucket=stage_uri.netloc,
                    blob=blob,
                ),
                timestamp_units=TimestampUnit.MICROS,
            ),
        )
    else:
        msg = f"Unexpected staging uri scheme: {stage_uri.scheme}"
        raise NotImplementedError(msg)
    run_online_store_copier(request)

    # Issue status update
    if fd.is_temporal_aggregate and not fd.is_continuous:
        anchor_time = feature_end_timestamp.ToDatetime() - fd.aggregate_slide_interval.ToTimedelta()
        anchor_time_pb = timestamp_pb2.Timestamp()
        anchor_time_pb.FromDatetime(anchor_time)
        status_update = StatusUpdateRequest(anchor_time=anchor_time_pb)
    else:
        status_update = StatusUpdateRequest(materialized_raw_data_end_time=feature_end_timestamp)

    status_request = OnlineStoreCopierRequest(
        online_store_writer_configuration=online_store_writer_config,
        feature_view=feature_view,
        status_update_request=status_update,
    )
    run_online_store_copier(status_request)


def get_delta_writer(
    params: MaterializationTaskParams,
    store_params_override: Optional[OfflineStoreParams] = None,
    table_uri_override: Optional[str] = None,
) -> DeltaWriter:
    role = params.dynamodb_cross_account_role if params.HasField("dynamodb_cross_account_role") else None
    kms_key_arn = params.kms_key_arn if params.HasField("kms_key_arn") else None
    fd = feature_definition_from_task_params(params)

    return DeltaWriter(
        store_params=store_params_override if store_params_override else OfflineStoreParams.for_feature_definition(fd),
        table_uri=table_uri_override if table_uri_override else params.offline_store_path,
        join_keys=fd.join_keys,
        dynamodb_log_table_name=params.delta_log_table,
        dynamodb_log_table_region=params.dynamodb_table_region,
        dynamodb_cross_account_role=role,
        kms_key_arn=kms_key_arn,
    )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/common/compute_mode__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.tecton_proto/common/compute_mode__client.proto\x12\x13tecton_proto.common*\x93\x01\n\x10\x42\x61tchComputeMode\x12\"\n\x1e\x42\x41TCH_COMPUTE_MODE_UNSPECIFIED\x10\x00\x12\x1c\n\x18\x42\x41TCH_COMPUTE_MODE_SPARK\x10\x01\x12 \n\x1c\x42\x41TCH_COMPUTE_MODE_SNOWFLAKE\x10\x02\x12\x1b\n\x17\x42\x41TCH_COMPUTE_MODE_RIFT\x10\x03\x42\x15\n\x11\x63om.tecton.commonP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.common.compute_mode__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\021com.tecton.commonP\001'
  _BATCHCOMPUTEMODE._serialized_start=72
  _BATCHCOMPUTEMODE._serialized_end=219
# @@protoc_insertion_point(module_scope)



from unittest import TestCase
from unittest import mock

from tecton._internals.metadata_service_impl.auth_lib import InternalAuthProvider
from tecton._internals.metadata_service_impl.request_lib import InternalRequestProvider
from tecton_core import conf
from tecton_core.metadata_service_impl import trace


@mock.patch("tecton.identities.okta.get_token_refresh_if_needed", lambda: None)
class RequestLibTest(TestCase):
    @classmethod
    def setUp(cls):
        conf._CONFIG_OVERRIDES = {}
        conf.set("API_SERVICE", "https://acme.tecton.ai/api")

    def test_okta_token(self):
        with mock.patch("tecton.identities.okta.get_token_refresh_if_needed", lambda: "abc123"):
            auth_provider = InternalAuthProvider()
            headers = InternalRequestProvider(auth_provider).request_headers()
            assert headers["authorization"] == "Bearer abc123"
            assert auth_provider.get_auth_header() is not None

    def test_api_token(self):
        conf.set("TECTON_API_KEY", "my_token22")
        auth_provider = InternalAuthProvider()
        headers = InternalRequestProvider(auth_provider).request_headers()
        assert headers["authorization"] == "Tecton-key my_token22"

    def test_no_token(self):
        assert InternalAuthProvider().get_auth_header() is None

    def test_workspace(self):
        conf.set("TECTON_WORKSPACE", "my_ws")
        headers = InternalRequestProvider(InternalAuthProvider()).request_headers()
        assert headers["x-workspace"] == "my_ws"
        assert "x-tecton-force-emr" not in headers

    def test_workspace_emr_override(self):
        conf.set("TECTON_WORKSPACE", "my_ws__emr")
        headers = InternalRequestProvider(InternalAuthProvider()).request_headers()
        assert headers["x-workspace"] == "my_ws__emr"
        assert headers["x-tecton-force-emr"] == "true"

    def test_trace_id(self):
        trace.set_trace_id("traceme_123")
        headers = InternalRequestProvider(InternalAuthProvider()).request_headers()
        assert headers["x-trace-id"] == "traceme_123"

    def test_version(self):
        with mock.patch("tecton.version.get_semantic_version", lambda: "1.2.3"):
            headers = InternalRequestProvider(InternalAuthProvider()).request_headers()
            assert headers["x-tecton-client-version"] == "1.2.3"

    def test_url(self):
        conf.set("API_SERVICE", "https://acme.tecton.ai/api")
        request_provider = InternalRequestProvider(InternalAuthProvider())
        assert request_provider.request_url() == "https://acme.tecton.ai/api/proxy"
        assert request_provider.request_headers()["host"] == "acme.tecton.ai"

    def test_url_port(self):
        conf.set("API_SERVICE", "https://acme.tecton.ai:5000/api")
        request_provider = InternalRequestProvider(InternalAuthProvider())
        assert request_provider.request_url() == "https://acme.tecton.ai:5000/api/proxy"
        assert request_provider.request_headers()["host"] == "acme.tecton.ai:5000"



import logging
import time
from typing import Optional

from pyspark.sql import SparkSession

from tecton_core.query_consts import anchor_time
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.materialization_utils import get_statsd_client
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark.offline_store import OfflineStoreWriterParams
from tecton_spark.offline_store import get_offline_store_writer


logger = logging.getLogger(__name__)


def _export_deletion_metrics(
    spark: SparkSession,
    store_type: str,
    metric_source: str,
    requested_num_keys: int,
    deleted_num_keys: int,
    latency: Optional[float] = None,
):
    statsd_client = get_statsd_client(spark)

    statsd_client.incr(f"{metric_source}.{store_type}.num_keys_deletion_requested", requested_num_keys)
    statsd_client.incr(f"{metric_source}.{store_type}.num_keys_deletion_success", deleted_num_keys)

    if latency:
        # timing expects number of milliseconds
        statsd_client.timing(f"{metric_source}.{store_type}.deletion_latency", latency * 1000)


def run_online_store_deleter(spark: SparkSession, materialization_task_params: MaterializationTaskParams):
    start = time.time()
    report = spark._jvm.com.tecton.onlinestorewriter.deleter.OnlineStoreDeleter.fromMaterializationTaskParams(
        materialization_task_params.SerializeToString()
    ).run()
    latency = time.time() - start

    if materialization_task_params.HasField("online_store_writer_config"):
        store_type = materialization_task_params.online_store_writer_config.online_store_params.WhichOneof("store_type")
    else:
        store_type = "dynamo"

    _export_deletion_metrics(
        spark,
        store_type,
        metric_source="tecton-online-store-deleter",
        requested_num_keys=report.getRequestedNumKeys(),
        deleted_num_keys=report.getDeletedNumKeys(),
        latency=latency,
    )

    if report.getError():
        msg = f"Deletion was interrupted due to the error received from storage: {report.getError()}"
        raise RuntimeError(msg)


def run_offline_store_deleter(spark: SparkSession, materialization_task_params: MaterializationTaskParams):
    deletion_params = materialization_task_params.deletion_task_info.deletion_parameters
    spark.conf.set(
        "spark.databricks.delta.commitInfo.userMetadata",
        f'{{"deletionPath":"{deletion_params.offline_join_keys_path}"}}',
    )
    fd = feature_definition_from_task_params(materialization_task_params)
    if fd.is_temporal_aggregate:
        time_column = anchor_time()
    else:
        time_column = fd.time_key
    offline_store_params = OfflineStoreWriterParams(
        s3_path=materialization_task_params.offline_store_path,
        always_store_anchor_column=False,
        time_column=time_column,
        join_key_columns=fd.join_keys,
        is_continuous=False,
    )
    store_writer = get_offline_store_writer(offline_store_params, fd, fd.get_feature_store_format_version, spark)
    keys = spark.read.parquet(deletion_params.offline_join_keys_path)
    keys = keys.distinct()
    requested_num_keys = keys.count()
    start = time.time()
    deleted_num_keys = store_writer.delete_keys(keys)
    latency = time.time() - start

    store_type = fd.offline_store_config.WhichOneof("store_type") if fd.offline_store_config else None
    store_type = store_type or "delta"

    _export_deletion_metrics(
        spark,
        store_type=store_type,
        metric_source="tecton-offline-store-deleter",
        requested_num_keys=requested_num_keys,
        deleted_num_keys=deleted_num_keys,
        latency=latency,
    )



from abc import ABC
from datetime import timedelta
from typing import Optional

from tecton._internals.display import Displayable
from tecton.framework.dataset import SavedDataset


class TectonJob(ABC):
    def cancel_job(self) -> bool:
        pass

    # wait for all sub-tasks to complete
    def wait_for_completion(self, timeout: Optional[timedelta] = None):
        pass

    def get_job_status_for_display(self) -> Displayable:
        pass


class DatasetJob(TectonJob):
    _dataset: SavedDataset

    # cancel all sub-tasks
    def cancel_job(self) -> bool:
        pass

    # wait for all sub-tasks to complete
    def wait_for_completion(self, timeout: Optional[timedelta] = None):
        pass

    # return dataset object, if job is completed
    def get_dataset(self) -> Optional[SavedDataset]:
        return self._dataset

    def get_job_status_for_display(self) -> Displayable:
        pass



import datetime
import logging

import pendulum
from pyspark.sql import DataFrame
from pyspark.sql import SparkSession
from pyspark.sql.utils import AnalysisException

from tecton_core import specs
from tecton_core.fco_container import FcoContainer
from tecton_core.fco_container import create_fco_container
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.id_helper import IdHelper
from tecton_core.offline_store import PartitionType
from tecton_core.query_consts import valid_from
from tecton_core.time_utils import convert_to_effective_timestamp
from tecton_materialization.batch_materialization import DEFAULT_COALESCE_FOR_S3
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_spark.offline_store import DeltaWriter
from tecton_spark.offline_store import OfflineStoreWriterParams


logger = logging.getLogger(__name__)


def feature_export_from_params(spark: SparkSession, task_params: MaterializationTaskParams):
    export_params = task_params.feature_export_info.feature_export_parameters
    start_time = export_params.feature_start_time.ToDatetime()
    end_time = export_params.feature_end_time.ToDatetime()
    feature_view_id_str = IdHelper.to_string(task_params.feature_view.feature_view_id)
    parent_materialization_id_str = IdHelper.to_string(export_params.parent_materialization_task_id)

    logger.info(
        f"Starting feature export {task_params.materialization_task_id} job for feature view: {feature_view_id_str}, parent_materialization_task_id: {parent_materialization_id_str} for time range {start_time} to {end_time}"
    )
    fco_container = create_fco_container(
        list(task_params.virtual_data_sources) + list(task_params.transformations) + list(task_params.entities),
        deserialize_funcs_to_main=True,
    )
    fv_spec = specs.create_feature_view_spec_from_data_proto(task_params.feature_view)
    fd = FeatureDefinitionWrapper(fv_spec, fco_container)

    spark_df, effective_tile = _get_features_in_range(fv_spec, fco_container, start_time, end_time)

    if spark_df.rdd.isEmpty():
        logging.info(
            f"No features found for materialization time range {start_time} to {end_time} and effective time range {effective_tile.start} to {effective_tile.end}"
        )
        return

    spark.conf.set(
        "spark.databricks.delta.commitInfo.userMetadata",
        f'{{"featureStartTime":"{export_params.feature_start_time.ToJsonString()}", "featureEndTime": "{export_params.feature_end_time.ToJsonString()}"}}',
    )
    is_write_optimized = spark.conf.get("spark.databricks.delta.optimizeWrite.enabled", None) == "true"
    df_to_write = spark_df if is_write_optimized else spark_df.coalesce(DEFAULT_COALESCE_FOR_S3)

    table_location = export_params.export_store_path
    partition_size = datetime.timedelta(days=1)
    version = fd.get_feature_store_format_version
    logging.info(
        f"Writing features to {table_location} for materialization time range {start_time} to {end_time} and effective time range {effective_tile.start} to {effective_tile.end} with partition size={partition_size}"
    )

    export_store_params = OfflineStoreWriterParams(
        s3_path=table_location,
        always_store_anchor_column=True,
        time_column=valid_from(),
        join_key_columns=fd.join_keys,
        is_continuous=fd.is_continuous,
    )

    writer = DeltaWriter(export_store_params, spark, version, partition_size, PartitionType.DATE_STR)
    writer.overwrite_dataframe_in_tile(df_to_write, effective_tile)


def _get_features_in_range(
    fv_spec: specs.FeatureViewSpec,
    fco_container: FcoContainer,
    materialization_start_time: datetime.datetime,
    materialization_end_time: datetime.datetime,
) -> (DataFrame, pendulum.Period):
    from tecton.framework.feature_view import feature_view_from_spec

    fdw = FeatureDefinitionWrapper(fv_spec, fco_container)
    fv = feature_view_from_spec(fv_spec, fco_container)

    # get_features_in_range queries features based on their effective timestamp
    batch_schedule_seconds = fdw.batch_materialization_schedule.in_seconds()
    data_delay = fdw.online_store_data_delay_seconds
    effective_start_time = datetime.datetime.fromtimestamp(
        convert_to_effective_timestamp(int(materialization_start_time.timestamp()), batch_schedule_seconds, data_delay)
    )
    effective_end_time = datetime.datetime.fromtimestamp(
        convert_to_effective_timestamp(int(materialization_end_time.timestamp()), batch_schedule_seconds, data_delay)
    )

    tecton_df = fv.get_features_in_range(
        start_time=effective_start_time, end_time=effective_end_time, from_source=False
    )

    try:
        df = tecton_df.to_spark()
    except AnalysisException as e:
        # TODO: handle empty stores better instead of relying on string matching on an error
        if "Unable to infer schema for Parquet. It must be specified manually." in str(e):
            logger.warning("Unable to infer Parquet schema; assuming staging offline store is empty")
            return
        else:
            raise

    return df, pendulum.period(effective_start_time, effective_end_time)



from typing import Optional

import click

from tecton import conf
from tecton._internals.utils import cluster_url
from tecton.cli import cli_utils
from tecton.cli import printer
from tecton.cli.command import TectonCommand
from tecton.identities import credentials
from tecton.identities import okta


@click.command(requires_auth=False, cls=TectonCommand)
@click.argument("tecton_url", required=False)
@click.option(
    "--manual/--no-manual",
    default=False,
    help="Manually require user to open browser and paste login token. Needed when using the Tecton CLI in a headless environment.",
)
@click.option("--okta-session-token", default=None, hidden=True, required=False)
def login(tecton_url: Optional[str], manual: bool, okta_session_token: Optional[str]):
    """Log in and authenticate Tecton CLI.

    The Tecton URL may be optionally passed on the command line as TECTON_URL, otherwise you will be prompted.
    """
    host = cluster_url()

    if tecton_url is None:
        printer.safe_print("Enter configuration. Press enter to use current value")
        prompt = "Tecton Cluster URL [%s]: " % (host or "no current value. example: https://yourco.tecton.ai")
        new_host = input(prompt).strip()
        if new_host:
            host = new_host
    else:
        host = tecton_url

    if okta_session_token:
        auth_flow_type = okta.AuthFlowType.SESSION_TOKEN
    elif manual:
        auth_flow_type = okta.AuthFlowType.BROWSER_MANUAL
    else:
        auth_flow_type = okta.AuthFlowType.BROWSER_HANDS_FREE

    temp_result, auth_code = credentials._initiate_login_and_get_auth_code(
        host=host, auth_flow_type=auth_flow_type, okta_session_token=okta_session_token
    )
    credentials._complete_login_with_auth_code(temp_result, auth_code)
    printer.safe_print(f"✅ Updated configuration at {conf._LOCAL_TECTON_CONFIG_FILE}")


@click.command(requires_auth=False, cls=TectonCommand)
def logout():
    """Log out of current user session. This is a no-op if there is no logged in session."""
    conf.delete_okta_tokens()
    printer.safe_print("✅ Logged user out of session. Use `tecton login` to re-authenticate.")


@click.command(cls=TectonCommand)
def whoami():
    """Show the current User or API Key used to authenticate with Tecton."""
    profile = credentials.who_am_i()
    if profile and isinstance(profile, okta.UserProfile):
        key_map = {"id": "ID", "email": "Email", "name": "Name"}
        cli_utils.pprint_attr_obj(key_map, profile, colwidth=16)
    elif profile and isinstance(profile, credentials.ServiceAccountProfile):
        service_account = {
            "Service Account ID": profile.id,
            "Secret Key": profile.obscured_key,
            "Name": profile.name,
            "Description": profile.description,
            "Created by": profile.created_by,
        }
        cli_utils.pprint_dict(service_account, colwidth=19)
    else:
        printer.safe_print(
            "Tecton credentials are not configured or have expired. Run `tecton login` or set an "
            "API Key to authenticate"
        )



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/transformation__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from tecton_proto.args import transformation__client_pb2 as tecton__proto_dot_args_dot_transformation____client__pb2
from tecton_proto.args import user_defined_function__client_pb2 as tecton__proto_dot_args_dot_user__defined__function____client__pb2
from tecton_proto.common import id__client_pb2 as tecton__proto_dot_common_dot_id____client__pb2
from tecton_proto.data import fco_metadata__client_pb2 as tecton__proto_dot_data_dot_fco__metadata____client__pb2
from tecton_proto.validation import validator__client_pb2 as tecton__proto_dot_validation_dot_validator____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n.tecton_proto/data/transformation__client.proto\x12\x11tecton_proto.data\x1a.tecton_proto/args/transformation__client.proto\x1a\x35tecton_proto/args/user_defined_function__client.proto\x1a$tecton_proto/common/id__client.proto\x1a,tecton_proto/data/fco_metadata__client.proto\x1a/tecton_proto/validation/validator__client.proto\"\xa4\x04\n\x0eTransformation\x12\x44\n\x11transformation_id\x18\x01 \x01(\x0b\x32\x17.tecton_proto.common.IdR\x10transformationId\x12\x41\n\x0c\x66\x63o_metadata\x18\x02 \x01(\x0b\x32\x1e.tecton_proto.data.FcoMetadataR\x0b\x66\x63oMetadata\x12V\n\x13transformation_mode\x18\x03 \x01(\x0e\x32%.tecton_proto.args.TransformationModeR\x12transformationMode\x12K\n\ruser_function\x18\x04 \x01(\x0b\x32&.tecton_proto.args.UserDefinedFunctionR\x0cuserFunction\x12^\n\x0fvalidation_args\x18\x05 \x01(\x0b\x32\x35.tecton_proto.validation.TransformationValidationArgsR\x0evalidationArgs\x12H\n\x07options\x18\x06 \x03(\x0b\x32..tecton_proto.data.Transformation.OptionsEntryR\x07options\x1a:\n\x0cOptionsEntry\x12\x10\n\x03key\x18\x01 \x01(\tR\x03key\x12\x14\n\x05value\x18\x02 \x01(\tR\x05value:\x02\x38\x01\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.transformation__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _TRANSFORMATION_OPTIONSENTRY._options = None
  _TRANSFORMATION_OPTIONSENTRY._serialized_options = b'8\001'
  _TRANSFORMATION._serialized_start=306
  _TRANSFORMATION._serialized_end=854
  _TRANSFORMATION_OPTIONSENTRY._serialized_start=796
  _TRANSFORMATION_OPTIONSENTRY._serialized_end=854
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/remote_compute_environment__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from tecton_proto.common import container_image__client_pb2 as tecton__proto_dot_common_dot_container__image____client__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n:tecton_proto/data/remote_compute_environment__client.proto\x12\x11tecton_proto.data\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x31tecton_proto/common/container_image__client.proto\"\xa0\x08\n\x18RemoteComputeEnvironment\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x38\n\x04type\x18\x03 \x01(\x0e\x32$.tecton_proto.data.RemoteComputeTypeR\x04type\x12\x42\n\x06status\x18\x04 \x01(\x0e\x32*.tecton_proto.data.RemoteEnvironmentStatusR\x06status\x12\x42\n\nimage_info\x18\x05 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\timageInfo\x12\x39\n\ncreated_at\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tcreatedAt\x12\x39\n\nupdated_at\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tupdatedAt\x12 \n\x0b\x64\x65scription\x18\t \x01(\tR\x0b\x64\x65scription\x12%\n\x0epython_version\x18\x0b \x01(\tR\rpythonVersion\x12\"\n\x0crequirements\x18\x0c \x01(\tR\x0crequirements\x12\x33\n\x15resolved_requirements\x18\r \x01(\tR\x14resolvedRequirements\x12,\n\x12s3_wheels_location\x18\x0e \x01(\tR\x10s3WheelsLocation\x12U\n\x10\x66\x65\x61ture_services\x18\x0f \x03(\x0b\x32*.tecton_proto.data.DependentFeatureServiceR\x0f\x66\x65\x61tureServices\x12`\n\x18realtime_job_environment\x18\x11 \x01(\x0b\x32&.tecton_proto.data.RealtimeEnvironmentR\x16realtimeJobEnvironment\x12\x64\n\x1arift_batch_job_environment\x18\x12 \x01(\x0b\x32\'.tecton_proto.data.RiftBatchEnvironmentR\x17riftBatchJobEnvironment\x12_\n\x1asupported_job_environments\x18\x13 \x03(\x0e\x32!.tecton_proto.data.JobEnvironmentR\x18supportedJobEnvironments\x12\x1f\n\x0bsdk_version\x18\x14 \x01(\tR\nsdkVersion\x12%\n\x0estatus_details\x18\x15 \x01(\tR\rstatusDetailsJ\x04\x08\x06\x10\x07J\x04\x08\n\x10\x0bJ\x04\x08\x10\x10\x11\"\x84\x01\n\x18\x41WSComputeInstanceGroups\x12\x32\n\x15\x61utoscaling_group_arn\x18\x01 \x01(\tR\x13\x61utoscalingGroupArn\x12\x34\n\x16\x61utoscaling_group_name\x18\x02 \x01(\tR\x14\x61utoscalingGroupName\"\xdd\x03\n\x13RealtimeEnvironment\x12G\n tecton_transform_runtime_version\x18\x01 \x01(\tR\x1dtectonTransformRuntimeVersion\x12\x42\n\nimage_info\x18\x02 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\timageInfo\x12\x30\n\x13remote_function_uri\x18\x03 \x01(\tH\x00R\x11remoteFunctionUri\x12l\n\x1b\x61ws_compute_instance_groups\x18\x06 \x01(\x0b\x32+.tecton_proto.data.AWSComputeInstanceGroupsH\x00R\x18\x61wsComputeInstanceGroups\x12U\n\x10\x66\x65\x61ture_services\x18\x05 \x03(\x0b\x32*.tecton_proto.data.DependentFeatureServiceR\x0f\x66\x65\x61tureServices\x12-\n\x12online_provisioned\x18\x07 \x01(\x08R\x11onlineProvisionedB\r\n\x0b\x64\x65stinationJ\x04\x08\x04\x10\x05\"\xf0\x01\n\x14RiftBatchEnvironment\x12S\n&tecton_materialization_runtime_version\x18\x01 \x01(\tR#tectonMaterializationRuntimeVersion\x12\x42\n\nimage_info\x18\x02 \x01(\x0b\x32#.tecton_proto.common.ContainerImageR\timageInfo\x12?\n\x1c\x63luster_environment_build_id\x18\x03 \x01(\tR\x19\x63lusterEnvironmentBuildId\"\x9c\x01\n\x1bRemoteEnvironmentUploadInfo\x12%\n\x0e\x65nvironment_id\x18\x01 \x01(\tR\renvironmentId\x12G\n\x0es3_upload_info\x18\x02 \x01(\x0b\x32\x1f.tecton_proto.data.S3UploadInfoH\x00R\x0cs3UploadInfoB\r\n\x0bupload_info\"o\n\x15ObjectStoreUploadPart\x12G\n\x0es3_upload_part\x18\x01 \x01(\x0b\x32\x1f.tecton_proto.data.S3UploadPartH\x00R\x0cs3UploadPartB\r\n\x0bupload_part\"o\n\x0cS3UploadInfo\x12\x1b\n\tupload_id\x18\x01 \x01(\tR\x08uploadId\x12\x42\n\x0cupload_parts\x18\x02 \x03(\x0b\x32\x1f.tecton_proto.data.S3UploadPartR\x0buploadParts\"\x8d\x01\n\x0cS3UploadPart\x12(\n\x10parent_upload_id\x18\x01 \x01(\tR\x0eparentUploadId\x12\x1f\n\x0bpart_number\x18\x02 \x01(\x05R\npartNumber\x12\x13\n\x05\x65_tag\x18\x03 \x01(\tR\x04\x65Tag\x12\x1d\n\nupload_url\x18\x04 \x01(\tR\tuploadUrl\"r\n\x17\x44\x65pendentFeatureService\x12%\n\x0eworkspace_name\x18\x01 \x01(\tR\rworkspaceName\x12\x30\n\x14\x66\x65\x61ture_service_name\x18\x02 \x01(\tR\x12\x66\x65\x61tureServiceName\"i\n\x14\x44\x65pendentFeatureView\x12%\n\x0eworkspace_name\x18\x01 \x01(\tR\rworkspaceName\x12*\n\x11\x66\x65\x61ture_view_name\x18\x02 \x01(\tR\x0f\x66\x65\x61tureViewName*\x90\x01\n\x0eJobEnvironment\x12\x1f\n\x1bJOB_ENVIRONMENT_UNSPECIFIED\x10\x00\x12\x1c\n\x18JOB_ENVIRONMENT_REALTIME\x10\x01\x12\x1e\n\x1aJOB_ENVIRONMENT_RIFT_BATCH\x10\x02\x12\x1f\n\x1bJOB_ENVIRONMENT_RIFT_STREAM\x10\x03*\xe1\x01\n\x17RemoteEnvironmentStatus\x12%\n!REMOTE_ENVIRONMENT_STATUS_PENDING\x10\x00\x12#\n\x1fREMOTE_ENVIRONMENT_STATUS_READY\x10\x01\x12#\n\x1fREMOTE_ENVIRONMENT_STATUS_ERROR\x10\x02\x12&\n\"REMOTE_ENVIRONMENT_STATUS_DELETING\x10\x03\x12-\n)REMOTE_ENVIRONMENT_STATUS_DELETION_FAILED\x10\x04*\xab\x01\n\x11RemoteComputeType\x12\x1c\n\x18REMOTE_COMPUTE_TYPE_CORE\x10\x00\x12 \n\x1cREMOTE_COMPUTE_TYPE_EXTENDED\x10\x01\x12\x36\n2REMOTE_COMPUTE_TYPE_SNOWPARK_DEPRECATED_DO_NOT_USE\x10\x02\x12\x1e\n\x1aREMOTE_COMPUTE_TYPE_CUSTOM\x10\x03\x42\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.remote_compute_environment__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _JOBENVIRONMENT._serialized_start=2835
  _JOBENVIRONMENT._serialized_end=2979
  _REMOTEENVIRONMENTSTATUS._serialized_start=2982
  _REMOTEENVIRONMENTSTATUS._serialized_end=3207
  _REMOTECOMPUTETYPE._serialized_start=3210
  _REMOTECOMPUTETYPE._serialized_end=3381
  _REMOTECOMPUTEENVIRONMENT._serialized_start=166
  _REMOTECOMPUTEENVIRONMENT._serialized_end=1222
  _AWSCOMPUTEINSTANCEGROUPS._serialized_start=1225
  _AWSCOMPUTEINSTANCEGROUPS._serialized_end=1357
  _REALTIMEENVIRONMENT._serialized_start=1360
  _REALTIMEENVIRONMENT._serialized_end=1837
  _RIFTBATCHENVIRONMENT._serialized_start=1840
  _RIFTBATCHENVIRONMENT._serialized_end=2080
  _REMOTEENVIRONMENTUPLOADINFO._serialized_start=2083
  _REMOTEENVIRONMENTUPLOADINFO._serialized_end=2239
  _OBJECTSTOREUPLOADPART._serialized_start=2241
  _OBJECTSTOREUPLOADPART._serialized_end=2352
  _S3UPLOADINFO._serialized_start=2354
  _S3UPLOADINFO._serialized_end=2465
  _S3UPLOADPART._serialized_start=2468
  _S3UPLOADPART._serialized_end=2609
  _DEPENDENTFEATURESERVICE._serialized_start=2611
  _DEPENDENTFEATURESERVICE._serialized_end=2725
  _DEPENDENTFEATUREVIEW._serialized_start=2727
  _DEPENDENTFEATUREVIEW._serialized_end=2832
# @@protoc_insertion_point(module_scope)



import sys

import click

from tecton import tecton_context
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton._internals.utils import format_freshness_table
from tecton._internals.utils import format_materialization_attempts
from tecton._internals.utils import get_all_freshness
from tecton.cli import printer
from tecton.cli import workspace_utils
from tecton.cli.command import TectonCommand
from tecton.cli.workspace_utils import WorkspaceType
from tecton_core.fco_container import FcoContainer
from tecton_core.id_helper import IdHelper
from tecton_proto.metadataservice import metadata_service__client_pb2 as metadata_service_pb2


@click.command(cls=TectonCommand)
@click.argument("feature_view_name")
@click.option("--limit", default=100, type=int, help="Set the maximum limit of results.")
@click.option("--errors-only/--no-errors-only", default=False, help="Only show errors.")
@click.option(
    "--workspace",
    default=None,
    type=WorkspaceType(),
    help="Name of the workspace containing FEATURE_VIEW_NAME. Defaults to the currently selected workspace.",
)
@click.option("--all-columns", is_flag=True, help="Display extra detail columns in the output table.")
def materialization_status(feature_view_name, limit, errors_only, workspace, all_columns):
    """Print materialization status for a specified Feature View in the current workspace."""
    # Fetch FeatureView
    workspace_name = workspace if workspace else tecton_context.get_current_workspace()
    workspace_utils.check_workspace_exists(workspace_name)
    fv_request = metadata_service_pb2.GetFeatureViewRequest(
        version_specifier=feature_view_name, workspace=workspace_name
    )
    fv_response = metadata_service.instance().GetFeatureView(fv_request)
    fco_container = FcoContainer.from_proto(fv_response.fco_container)
    fv_spec = fco_container.get_single_root()
    if fv_spec is None:
        printer.safe_print(f"Feature view '{feature_view_name}' not found.")
        sys.exit(1)
    fv_id = IdHelper.from_string(fv_spec.id)

    # Fetch Materialization Status
    status_request = metadata_service_pb2.GetMaterializationStatusRequest(
        feature_package_id=fv_id, workspace=workspace_name
    )
    status_response = metadata_service.instance().GetMaterializationStatus(status_request)

    column_names, materialization_status_rows = format_materialization_attempts(
        status_response.materialization_status.materialization_attempts,
        verbose=all_columns,
        limit=limit,
        errors_only=errors_only,
    )

    # Setting `max_width=0` creates a table with an unlimited width.
    table = Displayable.from_table(headings=column_names, rows=materialization_status_rows, max_width=0)
    # Align columns in the middle horizontally
    table._text_table.set_cols_align(["c" for _ in range(len(column_names))])
    printer.safe_print(table)


@click.command(cls=TectonCommand)
@click.option(
    "--workspace",
    default=None,
    type=WorkspaceType(),
    help="Name of the workspace to query. Defaults to the currently selected workspace.",
)
def freshness(workspace):
    """Print feature freshness for Feature Views in the current workspace."""
    # TODO: use GetAllFeatureFreshnessRequest once we implement Chronosphere based API.
    workspace_name = workspace if workspace else tecton_context.get_current_workspace()
    workspace_utils.check_workspace_exists(workspace_name)
    freshness_statuses = get_all_freshness(workspace_name)
    num_fvs = len(freshness_statuses)
    if num_fvs == 0:
        printer.safe_print("No Feature Views found in this workspace.")
        return

    printer.safe_print(format_freshness_table(freshness_statuses))



import typing
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import attrs
from typeguard import typechecked

from tecton_core import specs
from tecton_core.id_helper import IdHelper
from tecton_proto.common import id__client_pb2 as id_pb2
from tecton_proto.data import fco__client_pb2 as fco_pb2
from tecton_proto.data.entity__client_pb2 import Entity as EntityProto
from tecton_proto.data.feature_view__client_pb2 import FeatureView as FeatureViewProto
from tecton_proto.data.transformation__client_pb2 import Transformation
from tecton_proto.data.virtual_data_source__client_pb2 import VirtualDataSource as DataSourceProto


@attrs.frozen
class FcoContainer:
    """A wrapper class for FcoContainer proto, contains convenience accessors."""

    _root_ids: Tuple[str]
    _id_to_spec: Dict[str, specs.TectonObjectSpec]

    @classmethod
    @typechecked
    def from_proto(cls, proto: fco_pb2.FcoContainer, deserialize_funcs_to_main: bool = False) -> "FcoContainer":
        """Construct an FcoContainer from an FcoContainer proto."""
        root_ids = [IdHelper.to_string(id) for id in proto.root_ids]
        specs = [_spec_from_fco_data_proto(fco, deserialize_funcs_to_main) for fco in proto.fcos]
        id_to_spec = {spec.id: spec for spec in specs}
        return cls(root_ids=root_ids, id_to_spec=id_to_spec)  # type: ignore

    @classmethod
    @typechecked
    def from_specs(cls, specs: Sequence[specs.TectonObjectSpec], root_ids: Sequence[str]) -> "FcoContainer":
        """Construct an FcoContainer from a set of specs."""
        id_to_spec = {spec.id: spec for spec in specs}
        return cls(root_ids=tuple(root_ids), id_to_spec=id_to_spec)  # type: ignore

    @typechecked
    def get_by_id(self, id: str) -> specs.TectonObjectSpec:
        """
        :return: The TectonObjectSpec with the provided `id`.
        """
        return self._id_to_spec[id]

    @typechecked
    def get_by_id_proto(self, id_proto: id_pb2.Id) -> specs.TectonObjectSpec:
        """
        :return: The TectonObjectSpec with the provided `id` proto.
        """
        id_string = IdHelper.to_string(id_proto)
        return self._id_to_spec[id_string]

    @typechecked
    def get_by_ids(self, ids: Sequence[str]) -> List[specs.TectonObjectSpec]:
        """
        :return: The TectonObjectSpec with the provided `ids`.
        """

        return [self.get_by_id(id) for id in ids]

    @typechecked
    def get_single_root(self) -> Optional[specs.TectonObjectSpec]:
        """
        :return: The root TectonObjectSpec for the container or None. Errors if len(root_ids) > 1
        """

        num_root_ids = len(self._root_ids)
        if num_root_ids == 0:
            return None
        elif num_root_ids > 1:
            msg = f"Expected a single result but got {num_root_ids}"
            raise ValueError(msg)
        else:
            return self.get_by_id(self._root_ids[0])

    @typechecked
    def get_root_fcos(self) -> List[specs.TectonObjectSpec]:
        """
        :return: All root TectonObjectSpec for the container.
        """

        return [self.get_by_id(id) for id in self._root_ids]


FCO_CONTAINER_EMTPY = FcoContainer.from_proto(fco_pb2.FcoContainer())


def _spec_from_fco_data_proto(fco: fco_pb2.Fco, deserialize_funcs_to_main: bool) -> specs.TectonObjectSpec:
    if fco.HasField("virtual_data_source"):
        return specs.DataSourceSpec.from_data_proto(fco.virtual_data_source, deserialize_funcs_to_main)
    elif fco.HasField("entity"):
        return specs.EntitySpec.from_data_proto(fco.entity)
    elif fco.HasField("transformation"):
        return specs.TransformationSpec.from_data_proto(fco.transformation, deserialize_funcs_to_main)
    elif fco.HasField("feature_view"):
        return specs.create_feature_view_spec_from_data_proto(fco.feature_view)
    elif fco.HasField("feature_service"):
        return specs.FeatureServiceSpec.from_data_proto(fco.feature_service)
    else:
        msg = f"Unexpected fco type: {fco}"
        raise ValueError(msg)


DataProto = Union[DataSourceProto, Transformation, FeatureViewProto, EntityProto]


def _wrap_data_fco(inner_proto: DataProto) -> fco_pb2.Fco:
    fco = fco_pb2.Fco()
    if isinstance(inner_proto, DataSourceProto):
        fco.virtual_data_source.CopyFrom(inner_proto)
    elif isinstance(inner_proto, Transformation):
        fco.transformation.CopyFrom(inner_proto)
    elif isinstance(inner_proto, FeatureViewProto):
        fco.feature_view.CopyFrom(inner_proto)
    elif isinstance(inner_proto, EntityProto):
        fco.entity.CopyFrom(inner_proto)
    else:
        raise Exception("Unsupported type " + str(type(inner_proto)))
    return fco


def create_fco_container(
    fco_protos: typing.Iterable[DataProto], deserialize_funcs_to_main: bool = False
) -> FcoContainer:
    proto = fco_pb2.FcoContainer()
    for inner_fco_proto in fco_protos:
        wrapped_fco_proto = _wrap_data_fco(inner_fco_proto)
        proto.fcos.append(wrapped_fco_proto)
    return FcoContainer.from_proto(proto, deserialize_funcs_to_main=deserialize_funcs_to_main)



from datetime import datetime
from datetime import timezone
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pandas
import pendulum
from pyspark.sql import DataFrame as pysparkDF
from pyspark.sql.types import TimestampType

from tecton._internals import errors
from tecton._internals import metadata_service
from tecton._internals.display import Displayable
from tecton_core import specs
from tecton_core.errors import TectonValidationError
from tecton_core.fco_container import FcoContainer
from tecton_core.id_helper import IdHelper
from tecton_core.query.node_interface import DataframeWrapper
from tecton_core.query_consts import udf_internal
from tecton_proto.args.pipeline__client_pb2 import PipelineNode
from tecton_proto.data.freshness_status__client_pb2 import FreshnessStatus
from tecton_proto.data.materialization_status__client_pb2 import DataSourceType
from tecton_proto.data.materialization_status__client_pb2 import MaterializationAttemptStatus
from tecton_proto.data.materialization_status__client_pb2 import MaterializationStatusState
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetFeatureFreshnessRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import GetWorkspaceRequest
from tecton_proto.metadataservice.metadata_service__client_pb2 import QueryFeatureViewsRequest


_TIME_FORMAT = "%Y-%m-%d %H:%M:%S %Z"


def validate_join_keys(join_keys: List[str]):
    """
    Validates that `join_keys` is not empty and has non-empty distinct values.

    :raises TectonValidationError: if `join_keys` is invalid.
    """
    if not join_keys:
        msg = "join_keys"
        raise errors.EMPTY_ARGUMENT(msg)

    if "" in join_keys:
        msg = "join_keys"
        raise errors.EMPTY_ELEMENT_IN_ARGUMENT(msg)

    if len(join_keys) > len(set(join_keys)):
        msg = "join_keys"
        raise errors.DUPLICATED_ELEMENTS_IN_ARGUMENT(msg)


def validate_spine_dataframe(
    spine: DataframeWrapper, timestamp_key: Optional[str], request_context_keys: Optional[List[str]] = None
):
    spine_df = spine._dataframe
    if timestamp_key:
        if timestamp_key not in spine_df.columns:
            msg = "timestamp_key"
            raise errors.MISSING_SPINE_COLUMN(msg, timestamp_key, spine_df.columns)
        if isinstance(spine_df, pysparkDF):
            data_type = spine_df.schema[timestamp_key].dataType
            if not isinstance(data_type, TimestampType):
                raise errors.INVALID_SPINE_TIME_KEY_TYPE_SPARK(data_type)
        elif isinstance(spine_df, pandas.DataFrame):
            dtypes = dict(spine_df.dtypes)
            data_type = dtypes[timestamp_key]
            if not pandas.api.types.is_datetime64_any_dtype(data_type):
                raise errors.INVALID_SPINE_TIME_KEY_TYPE_PANDAS(data_type)
        else:
            import snowflake.snowpark
            from snowflake.snowpark import types

            if not isinstance(spine_df, snowflake.snowpark.DataFrame):
                for field in spine_df.schema.fields:
                    if field.name == timestamp_key and field.datatype != types.TimestampType():
                        raise errors.INVALID_SPINE_TIME_KEY_TYPE_SNOWFLAKE(field.datatype)
    if request_context_keys:
        for key in request_context_keys:
            if key not in spine_df.columns:
                raise errors.MISSING_REQUEST_DATA_IN_SPINE(key, spine_df.columns)


def format_seconds_into_highest_unit(total_seconds):
    intervals = [("wk", 60 * 60 * 24 * 7), ("d", 60 * 60 * 24), ("h", 60 * 60), ("m", 60), ("s", 1)]

    units = []
    for abbreviation, size in intervals:
        count = total_seconds // size
        total_seconds %= size
        if count != 0:
            units.append(f"{count}{abbreviation}")

    return " ".join(units[:2] if len(units) > 2 else units)


def snake_to_capitalized(snake_str):
    return "".join(x.title() for x in snake_str.split("_"))


def format_materialization_attempts(
    materialization_attempts, verbose=False, limit=1000, sort_columns=None, errors_only=False
) -> Tuple[List, List[List]]:
    """
    Formats a list of materialization attempts for use in Displayable.from_table
    Returns (column_names, materialization_status_rows).
    """
    column_names = ["TYPE", "WINDOW_START_TIME", "WINDOW_END_TIME", "STATUS", "ATTEMPT"]
    if verbose:
        column_names.extend(
            ["MATERIALIZATION_TASK", "ENV_VERSION", "TERMINATION_REASON", "STATE_MESSAGE", "TASK_SCHEDULED_AT"]
        )
    column_names.extend(["JOB_CREATED_AT", "JOB_LOGS"])

    materialization_attempts = materialization_attempts[:limit]
    if errors_only:
        materialization_attempts = [
            attempt
            for attempt in materialization_attempts
            if _materialization_status_state_name(attempt.materialization_state) == "ERROR"
        ]

    materialization_status_rows = []
    for attempt_status in materialization_attempts:
        data = _get_materialization_status_row_data(attempt_status)
        materialization_status_rows.append([data[c] for c in column_names])

    if sort_columns:
        keys = [k.upper() for k in sort_columns.split(",")]
        indices = []
        for key in keys:
            try:
                indices.append(column_names.index(key))
            except ValueError:
                msg = f"Unknown sort key {key}, should be one of: {','.join(column_names)}"
                raise ValueError(msg)
        materialization_status_rows.sort(key=lambda r: [r[i] for i in indices])

    return column_names, materialization_status_rows


def _materialization_status_state_name(state: MaterializationStatusState) -> str:
    state_name = MaterializationStatusState.Name(state)
    return state_name.replace("MATERIALIZATION_STATUS_STATE_", "")


def _get_materialization_status_row_data(attempt_status: MaterializationAttemptStatus) -> Dict[str, str]:
    status_dict = {}
    status_dict["TYPE"] = (
        "STREAM"
        if attempt_status.data_source_type == DataSourceType.DATA_SOURCE_TYPE_STREAM
        else "BATCH"
        if attempt_status.data_source_type == DataSourceType.DATA_SOURCE_TYPE_BATCH
        else "INGEST"
        if attempt_status.data_source_type == DataSourceType.DATA_SOURCE_TYPE_INGEST
        else "DELETION"
        if attempt_status.data_source_type == DataSourceType.DATA_SOURCE_TYPE_DELETION
        else "FEATURE PUBLISH"
        if attempt_status.data_source_type == DataSourceType.DATA_SOURCE_TYPE_FEATURE_EXPORT
        else "UNKNOWN"
    )
    status_dict["WINDOW_START_TIME"] = (
        "N/A"
        if not attempt_status.HasField("window_start_time")
        else attempt_status.window_start_time.ToDatetime().replace(tzinfo=timezone.utc).strftime(_TIME_FORMAT)
    )
    status_dict["WINDOW_END_TIME"] = (
        "N/A"
        if not attempt_status.HasField("window_end_time")
        else attempt_status.window_end_time.ToDatetime().replace(tzinfo=timezone.utc).strftime(_TIME_FORMAT)
    )
    status_dict["STATUS"] = _materialization_status_state_name(attempt_status.materialization_state)
    status_dict["ATTEMPT"] = "N/A" if not attempt_status.HasField("attempt_number") else attempt_status.attempt_number
    status_dict["MATERIALIZATION_TASK"] = IdHelper.to_string(attempt_status.materialization_task_id)
    status_dict["ENV_VERSION"] = attempt_status.spark_cluster_environment_version
    status_dict["TERMINATION_REASON"] = (
        "N/A"
        if not attempt_status.HasField("termination_reason")
        or len(attempt_status.termination_reason) == 0
        or attempt_status.termination_reason == "UNKNOWN_TERMINATION_REASON"
        else attempt_status.termination_reason
    )
    status_dict["STATE_MESSAGE"] = (
        "N/A"
        if not attempt_status.HasField("state_message") or len(attempt_status.state_message) == 0
        else attempt_status.state_message
    )
    status_dict["TASK_SCHEDULED_AT"] = (
        "N/A"
        if not attempt_status.HasField("materialization_task_created_at")
        else attempt_status.materialization_task_created_at.ToDatetime()
        .replace(tzinfo=timezone.utc)
        .strftime(_TIME_FORMAT)
    )
    status_dict["JOB_CREATED_AT"] = (
        "N/A"
        if not attempt_status.HasField("attempt_created_at")
        else attempt_status.attempt_created_at.ToDatetime().replace(tzinfo=timezone.utc).strftime(_TIME_FORMAT)
    )
    status_dict["JOB_LOGS"] = "N/A" if not attempt_status.HasField("run_page_url") else attempt_status.run_page_url

    return status_dict


def get_num_dependent_fv(node: PipelineNode, visited_inputs: Dict[str, bool]) -> int:
    if node.HasField("feature_view_node"):
        if node.feature_view_node.input_name in visited_inputs:
            return 0
        visited_inputs[node.feature_view_node.input_name] = True
        return 1
    elif node.HasField("transformation_node"):
        ret = 0
        for child in node.transformation_node.inputs:
            ret = ret + get_num_dependent_fv(child.node, visited_inputs)
        return ret
    return 0


def infer_timestamp(spine: Union[pandas.DataFrame, pysparkDF, DataframeWrapper]) -> Optional[str]:
    if isinstance(spine, DataframeWrapper):
        spine = spine._dataframe
    dtypes = dict(spine.dtypes)

    if isinstance(spine, pandas.DataFrame):
        timestamp_cols = [(k, v) for (k, v) in dtypes.items() if pandas.api.types.is_datetime64_any_dtype(v)]
    elif isinstance(spine, pysparkDF):
        timestamp_cols = [(k, v) for (k, v) in dtypes.items() if v == "timestamp"]
    else:
        msg = f"Unexpected data type for spine: {type(spine)}"
        raise TectonValidationError(msg)

    if len(timestamp_cols) > 1 or len(timestamp_cols) == 0:
        msg = f"Could not infer timestamp keys from {dtypes}; please specify explicitly"
        raise TectonValidationError(msg)
    return timestamp_cols[0][0]


def can_be_stale(ff_proto: FreshnessStatus) -> bool:
    return (
        ff_proto.expected_freshness.seconds > 0 and ff_proto.freshness.seconds > 0 and ff_proto.materialization_enabled
    )


def format_freshness_table(freshness_statuses: List[FreshnessStatus]) -> Displayable:
    timestamp_format = "%x %H:%M"
    headers = [
        "Feature View",
        "Materialized?",
        "Stale?",
        "Freshness",
        "Expected Freshness",
        "Created",
        "Stream?",
    ]

    freshness_data = [
        [
            ff_proto.feature_view_name,
            str(ff_proto.materialization_enabled),
            str(ff_proto.is_stale) if can_be_stale(ff_proto) else "-",
            format_seconds_into_highest_unit(ff_proto.freshness.seconds) if can_be_stale(ff_proto) else "-",
            format_seconds_into_highest_unit(ff_proto.expected_freshness.seconds) if can_be_stale(ff_proto) else "-",
            datetime.fromtimestamp(ff_proto.created_at.seconds).strftime(timestamp_format),
            str(ff_proto.is_stream),
        ]
        for ff_proto in freshness_statuses
    ]

    sort_order = {"True": 0, "False": 1, "-": 2}
    freshness_data = sorted(freshness_data, key=lambda row: sort_order[row[2]])
    table = Displayable.from_table(headings=headers, rows=freshness_data, max_width=0)
    table._text_table.set_cols_align(["c" for _ in range(len(headers))])
    return table


def get_all_freshness(workspace: str):
    request = QueryFeatureViewsRequest()
    request.workspace = workspace
    response = metadata_service.instance().QueryFeatureViews(request)
    fco_container = FcoContainer.from_proto(response.fco_container)
    fv_ids = []
    for spec in fco_container.get_root_fcos():
        if isinstance(spec, specs.MaterializedFeatureViewSpec):
            fv_ids.append(IdHelper.from_string(spec.id))

    freshness_statuses = []

    for fv_id in fv_ids:
        fresh_request = GetFeatureFreshnessRequest()
        fresh_request.fco_locator.id.CopyFrom(fv_id)
        fresh_request.fco_locator.workspace = workspace
        fresh_response = metadata_service.instance().GetFeatureFreshness(fresh_request)
        freshness_statuses.append(fresh_response.freshness_status)

    return freshness_statuses


def is_live_workspace(workspace_name: str) -> bool:
    request = GetWorkspaceRequest()
    request.workspace_name = workspace_name
    response = metadata_service.instance().GetWorkspace(request)
    return response.workspace.capabilities.materializable


def filter_internal_columns(df: pysparkDF) -> pysparkDF:
    output_columns = [f"`{c.name}`" for c in df.schema if udf_internal() not in c.name]
    return df.select(*output_columns)


def validate_request_data(
    request_data: Optional[dict],
    required_request_context_keys: List[str],
    is_read_api_v3=True,
):
    missing_keys = set(required_request_context_keys) - set(request_data.keys()) if request_data else None
    if missing_keys:
        # TODO:(read-api) remove this check and old error message once get_feature_vector is deleted
        if is_read_api_v3:
            raise errors.GET_ONLINE_FEATURES_MISSING_REQUEST_KEY(missing_keys)
        raise errors.GET_FEATURE_VECTOR_MISSING_REQUEST_KEY(missing_keys)


def get_time_limits_of_pandas_dataframe(df: pandas.DataFrame, time_key: str) -> Optional[pendulum.Period]:
    time_start = df[time_key].min()
    time_end = df[time_key].max()
    if pandas.isnull(time_start) or pandas.isnull(time_end):
        return None
    # Need to add 1 microsecond to the end time, since the range is exclusive at the end, and we need
    # to make sure to include the very last feature value (in terms of the event timestamp).
    return pendulum.instance(time_end).add(microseconds=1) - pendulum.instance(time_start)


def plural(x: int, singular: str, plural: str):
    """
    Returns the `singular` string if `x` is equal to 1 and `plural` if not.
    """
    if x == 1:
        return singular
    else:
        return plural


def cluster_url() -> Optional[str]:
    import sys

    from tecton_core import conf

    api_service = conf.get_or_none("API_SERVICE")
    if api_service:
        # API_SERVICE URLs of the form <subdomain>.tecton.ai/api are expected so this check
        # ensures an internal DNS address isn't being used or an invalid path is specified.
        if api_service.endswith("/api") and "ingress" not in api_service:
            return api_service[: -len("/api")]
        else:
            print(f"Warning: CLI is configured with non-standard URL: {api_service}", file=sys.stderr)
            return api_service
    else:
        return None



from tecton._internals.find_spark import find_spark as __find_spark


__initializing__ = True


__find_spark()

#  When adding/removing a class here, also mirror them in the tecton-docs repo.
#  We no longer autogenerate docs.


# ruff: noqa: F401 E402
from tecton import version as __version_lib
from tecton._internals.materialization_api import MaterializationAttemptData
from tecton._internals.materialization_api import MaterializationJobData
from tecton.fco_listers import list_data_sources
from tecton.fco_listers import list_datasets
from tecton.fco_listers import list_entities
from tecton.fco_listers import list_feature_services
from tecton.fco_listers import list_feature_tables
from tecton.fco_listers import list_feature_views
from tecton.fco_listers import list_transformations
from tecton.fco_listers import list_workspaces
from tecton.framework.configs import Aggregation
from tecton.framework.configs import BigQueryConfig
from tecton.framework.configs import BigtableConfig
from tecton.framework.configs import CacheConfig
from tecton.framework.configs import DatabricksClusterConfig
from tecton.framework.configs import DatabricksJsonClusterConfig
from tecton.framework.configs import DataprocJsonClusterConfig
from tecton.framework.configs import DatetimePartitionColumn
from tecton.framework.configs import DeltaConfig
from tecton.framework.configs import DynamoConfig
from tecton.framework.configs import EMRClusterConfig
from tecton.framework.configs import EMRJsonClusterConfig
from tecton.framework.configs import FileConfig
from tecton.framework.configs import HiveConfig
from tecton.framework.configs import KafkaConfig
from tecton.framework.configs import KafkaOutputStream
from tecton.framework.configs import KinesisConfig
from tecton.framework.configs import KinesisOutputStream
from tecton.framework.configs import LifetimeWindow
from tecton.framework.configs import LoggingConfig
from tecton.framework.configs import OfflineStoreConfig
from tecton.framework.configs import PandasBatchConfig
from tecton.framework.configs import ParquetConfig
from tecton.framework.configs import PushConfig
from tecton.framework.configs import RedisConfig
from tecton.framework.configs import RedshiftConfig
from tecton.framework.configs import RequestSource
from tecton.framework.configs import RiftBatchConfig
from tecton.framework.configs import Secret
from tecton.framework.configs import SnowflakeConfig
from tecton.framework.configs import SparkBatchConfig
from tecton.framework.configs import SparkStreamConfig
from tecton.framework.configs import TimeWindow
from tecton.framework.configs import TimeWindowSeries
from tecton.framework.configs import UnityCatalogAccessMode
from tecton.framework.configs import UnityConfig
from tecton.framework.configs import pandas_batch_config
from tecton.framework.configs import spark_batch_config
from tecton.framework.configs import spark_stream_config
from tecton.framework.data_frame import DataFrame
from tecton.framework.data_frame import FeatureVector
from tecton.framework.data_frame import TectonDataFrame
from tecton.framework.data_source import BatchSource
from tecton.framework.data_source import DataSource
from tecton.framework.data_source import PushSource
from tecton.framework.data_source import StreamSource
from tecton.framework.dataset import Dataset
from tecton.framework.entity import Entity
from tecton.framework.feature import Aggregate
from tecton.framework.feature import Attribute
from tecton.framework.feature import Embedding
from tecton.framework.feature import Inference
from tecton.framework.feature_service import FeatureService
from tecton.framework.feature_view import BatchFeatureView
from tecton.framework.feature_view import BatchTriggerType
from tecton.framework.feature_view import FeatureReference
from tecton.framework.feature_view import FeatureTable
from tecton.framework.feature_view import FeatureView
from tecton.framework.feature_view import MaterializedFeatureView
from tecton.framework.feature_view import RealtimeFeatureView
from tecton.framework.feature_view import StreamFeatureView
from tecton.framework.feature_view import StreamProcessingMode
from tecton.framework.feature_view import batch_feature_view
from tecton.framework.feature_view import on_demand_feature_view
from tecton.framework.feature_view import realtime_feature_view
from tecton.framework.feature_view import stream_feature_view
from tecton.framework.filtered_source import FilteredSource
from tecton.framework.model_config import ModelConfig
from tecton.framework.transformation import Transformation
from tecton.framework.transformation import const
from tecton.framework.transformation import transformation
from tecton.framework.utils import PANDAS_MODE
from tecton.framework.utils import PYSPARK_MODE
from tecton.framework.utils import PYTHON_MODE
from tecton.framework.utils import SNOWFLAKE_SQL_MODE
from tecton.framework.utils import SNOWPARK_MODE
from tecton.framework.utils import SPARK_SQL_MODE
from tecton.framework.validation_mode import ValidationMode
from tecton.framework.validation_mode import set_validation_mode
from tecton.framework.workspace import Workspace
from tecton.framework.workspace import get_data_source
from tecton.framework.workspace import get_entity
from tecton.framework.workspace import get_feature_service
from tecton.framework.workspace import get_feature_table
from tecton.framework.workspace import get_feature_view
from tecton.framework.workspace import get_transformation
from tecton.framework.workspace import get_workspace
from tecton.identities.credentials import clear_credentials
from tecton.identities.credentials import login
from tecton.identities.credentials import login_with_code
from tecton.identities.credentials import logout
from tecton.identities.credentials import set_credentials
from tecton.identities.credentials import test_credentials
from tecton.identities.credentials import who_am_i
from tecton.tecton_context import get_current_workspace
from tecton.tecton_context import set_tecton_spark_session
from tecton_core import conf
from tecton_core.compute_mode import ComputeMode
from tecton_core.filter_context import FilterContext
from tecton_core.materialization_context import materialization_context
from tecton_core.online_serving_index import OnlineServingIndex
from tecton_core.realtime_context import RealtimeContext


__version__ = __version_lib.get_version()
__initializing__ = False



import logging

from tecton._internals.sdk_decorators import sdk_public_method
from tecton_core import errors
from tecton_core.snowflake_context import SnowflakeContext


logger = logging.getLogger(__name__)


@sdk_public_method
def set_connection(connection) -> "SnowflakeContext":
    """
    Connect tecton to Snowflake.

    :param connection: The SnowflakeConnection object.
    :return: A SnowflakeContext object.
    """
    from snowflake.connector import SnowflakeConnection

    if not isinstance(connection, SnowflakeConnection):
        msg = "connection must be a SnowflakeConnection object"
        raise errors.TectonValidationError(msg)

    return SnowflakeContext.set_connection(connection)



from datetime import datetime
from io import BytesIO
from typing import Optional
from unittest import TestCase

import numpy
import pandas
import requests
from google.protobuf.struct_pb2 import Struct
from google.protobuf.struct_pb2 import Value
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal

from tecton._internals.query_helper import _QueryHelper
from tecton_core import errors
from tecton_proto.api.featureservice.feature_service__client_pb2 import FeatureServerComplexDataType
from tecton_proto.api.featureservice.feature_service__client_pb2 import FeatureServerDataType
from tecton_proto.api.featureservice.feature_service__client_pb2 import GetFeaturesResponse
from tecton_proto.api.featureservice.feature_service__client_pb2 import GetFeaturesResult
from tecton_proto.api.featureservice.feature_service__client_pb2 import Metadata
from tecton_proto.api.featureservice.feature_service__client_pb2 import QueryFeaturesResponse


def data_type(type_enum: int, element_type_enum: Optional[int] = None) -> FeatureServerComplexDataType:
    complex_type = FeatureServerComplexDataType()
    complex_type.type = type_enum
    if element_type_enum is not None:
        complex_type.element_type.type = element_type_enum
    return complex_type


def value_proto(value) -> Value:
    # Structs must be created from a top-level dictionary.
    wrapper_dict = {"k": value}
    s = Struct()
    s.update(wrapper_dict)
    return s.fields["k"]


class QueryHelperTest(TestCase):
    def setUp(self) -> None:
        self.query_helper = _QueryHelper("", feature_service_name="test1")

    def test_response_to_feature_vector(self):
        test_features = [
            ("f1", data_type(FeatureServerDataType.int64), value_proto("3")),
            ("f2", data_type(FeatureServerDataType.string), value_proto("three")),
            ("f3", data_type(FeatureServerDataType.float64), value_proto(33.3)),
            # The feature service JSON response uses strings for special float values (e.g. "NaN" and "Infinity").
            ("f4", data_type(FeatureServerDataType.float64), value_proto("Infinity")),
            ("f5", data_type(FeatureServerDataType.boolean), value_proto(True)),
            (
                "f6",
                data_type(FeatureServerDataType.array, FeatureServerDataType.string),
                value_proto(["one", "two", None]),
            ),
            ("f7", data_type(FeatureServerDataType.array, FeatureServerDataType.int64), value_proto(["1", "2", None])),
            (
                "f8",
                data_type(FeatureServerDataType.array, FeatureServerDataType.float32),
                value_proto([1.1, "Infinity", None]),
            ),
        ]
        response = GetFeaturesResponse()
        for f in test_features:
            feature = response.metadata.features.add()
            feature.name = f[0]
            feature.data_type.CopyFrom(f[1])
            response.result.features.extend([f[2]])

        actual_fv = self.query_helper._response_to_feature_vector(response, True, {"jk1": "abc"})
        actual_dict = actual_fv.to_dict()
        expected_dict = {
            "f1": 3,
            "f2": "three",
            "f3": 33.3,
            "f4": float("inf"),
            "f5": True,
            "f6": ["one", "two", None],
            "f7": [1, 2, None],
            "f8": [1.1, float("inf"), None],
            "jk1": "abc",
        }
        self.assertEqual(actual_dict, expected_dict)

        actual_pd = actual_fv.to_pandas()
        expected_pd = pandas.DataFrame(
            [
                [
                    "abc",
                    3,
                    "three",
                    33.3,
                    float("inf"),
                    True,
                    ["one", "two", None],
                    [1, 2, None],
                    [1.1, float("inf"), None],
                ]
            ],
            columns=["jk1", "f1", "f2", "f3", "f4", "f5", "f6", "f7", "f8"],
        )
        assert_frame_equal(actual_pd, expected_pd)

        actual_np = actual_fv.to_numpy()
        expected_np = numpy.array(
            ["abc", 3, "three", 33.3, float("inf"), True, ["one", "two", None], [1, 2, None], [1.1, float("inf"), None]]
        )
        assert_array_equal(actual_np, expected_np)

    def test_response_to_feature_vector_with_metadata(self):
        test_features = [
            ("f1", data_type(FeatureServerDataType.int64), value_proto("3"), 10000),
            ("f2", data_type(FeatureServerDataType.string), value_proto("three"), 100000),
            ("f3", data_type(FeatureServerDataType.float64), value_proto(33.3), 1000000),
            (
                "f4",
                data_type(FeatureServerDataType.array, FeatureServerDataType.string),
                value_proto(["one", "two", None]),
                10000000,
            ),
        ]
        response = GetFeaturesResponse()
        response.metadata.slo_info.slo_eligible = True
        for f in test_features:
            feature = response.metadata.features.add()
            feature.name = f[0]
            feature.data_type.CopyFrom(f[1])
            feature.effective_time.seconds = f[3]
            response.result.features.extend([f[2]])

        actual_fv = self.query_helper._response_to_feature_vector(response, True, {"jk1": "abc"})
        date1 = datetime.utcfromtimestamp(10000)
        date2 = datetime.utcfromtimestamp(100000)
        date3 = datetime.utcfromtimestamp(1000000)
        date4 = datetime.utcfromtimestamp(10000000)

        actual_dict = actual_fv.to_dict(return_effective_times=True)
        expected_dict = {
            "f1": {"value": 3, "effective_time": date1},
            "f2": {"value": "three", "effective_time": date2},
            "f3": {"value": 33.3, "effective_time": date3},
            "f4": {"value": ["one", "two", None], "effective_time": date4},
            "jk1": {"value": "abc", "effective_time": None},
        }
        self.assertEqual(actual_dict, expected_dict)

        actual_pd = actual_fv.to_pandas(return_effective_times=True)
        expected_pd = pandas.DataFrame(
            [
                ["jk1", "abc", None],
                ["f1", 3, date1],
                ["f2", "three", date2],
                ["f3", 33.3, date3],
                ["f4", ["one", "two", None], date4],
            ],
            columns=["name", "value", "effective_time"],
        )
        assert_frame_equal(actual_pd, expected_pd)

        actual_np = actual_fv.to_numpy(return_effective_times=True)
        expected_np = numpy.array([["abc", 3, "three", 33.3, ["one", "two", None]], [None, date1, date2, date3, date4]])
        assert_array_equal(actual_np, expected_np)

    def test_features_dict(self):
        result = GetFeaturesResult()
        meta = Metadata()

        feature = meta.features.add()
        feature.name = "race_track"
        feature.data_type.type = FeatureServerDataType.string
        result.features.extend([value_proto("silverstone")])

        wildcard_join_key = meta.join_keys.add()
        wildcard_join_key.name = "race_id"
        wildcard_join_key.data_type.type = FeatureServerDataType.int64
        result.join_keys.extend([value_proto("5")])

        actual = self.query_helper._feature_dict(result, meta)
        expected = {"race_track": "silverstone", "race_id": 5}
        self.assertEqual(actual, expected)

    def test_query_response_to_pandas(self):
        query_response = QueryFeaturesResponse()

        feature = query_response.metadata.features.add()
        feature.name = "race_track"
        feature.data_type.type = FeatureServerDataType.string
        wildcard_join_key = query_response.metadata.join_keys.add()
        wildcard_join_key.name = "race_id"
        wildcard_join_key.data_type.type = FeatureServerDataType.int64
        for i, val in enumerate([value_proto("silverstone"), value_proto("monza")]):
            result = query_response.results.add()
            feature_result = result.features.extend([val])
            join_key_result = result.join_keys.extend([value_proto(str(i))])

        actual = self.query_helper._query_response_to_pandas(query_response, {"race_season": 2020})
        expected = pandas.DataFrame(
            {
                "race_season": [2020, 2020],
                "race_track": ["silverstone", "monza"],
                "race_id": [0, 1],
            }
        )
        assert_frame_equal(actual, expected)

    def test_detailed_http_raise_for_status(self):
        def http_response(msg="OK", status=200, grpcCode=0):
            r = requests.Response()
            r.status_code = status
            r.raw = BytesIO(bytes(f'{{"message": "{msg}", "code": {grpcCode}}}', encoding="utf-8"))
            return r

        test_err_responses = [
            (http_response(msg="bad request", status=400), errors.TectonValidationError, "bad request"),
            (http_response(msg="no authN", status=401), errors.TectonValidationError, "no authN"),
            (http_response(msg="no authZ", status=403), errors.TectonValidationError, "no authZ"),
            (http_response(msg="not found", status=404), errors.TectonNotFoundError, "not found"),
            (http_response(msg="internal", status=500), errors.TectonInternalError, "internal"),
        ]

        # err responses
        for tr in test_err_responses:
            with self.assertRaisesRegex(tr[1], tr[2]) as e:
                self.query_helper._detailed_http_raise_for_status(tr[0])

        # ok response
        try:
            self.query_helper._detailed_http_raise_for_status(http_response())
        except Exception:
            self.assertTrue(False)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/args/data_source_config__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import duration_pb2 as google_dot_protobuf_dot_duration__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n2tecton_proto/args/data_source_config__client.proto\x12\x11tecton_proto.args\x1a\x1egoogle/protobuf/duration.proto\"\x13\n\x0b\x42\x61tchConfigJ\x04\x08\x01\x10\x02\"\xc7\x01\n\x0cStreamConfig\x12`\n\x17initial_stream_position\x18\x01 \x01(\x0e\x32(.tecton_proto.args.InitialStreamPositionR\x15initialStreamPosition\x12U\n\x19watermark_delay_threshold\x18\x02 \x01(\x0b\x32\x19.google.protobuf.DurationR\x17watermarkDelayThreshold*\x94\x01\n\x15InitialStreamPosition\x12\'\n#INITIAL_STREAM_POSITION_UNSPECIFIED\x10\x00\x12\"\n\x1eINITIAL_STREAM_POSITION_LATEST\x10\x01\x12(\n$INITIAL_STREAM_POSITION_TRIM_HORIZON\x10\x03\"\x04\x08\x02\x10\x02\x42\x13\n\x0f\x63om.tecton.argsP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.args.data_source_config__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.argsP\001'
  _INITIALSTREAMPOSITION._serialized_start=329
  _INITIALSTREAMPOSITION._serialized_end=477
  _BATCHCONFIG._serialized_start=105
  _BATCHCONFIG._serialized_end=124
  _STREAMCONFIG._serialized_start=127
  _STREAMCONFIG._serialized_end=326
# @@protoc_insertion_point(module_scope)



from abc import abstractmethod
from typing import List

import attrs
import pyspark

from tecton_core.query import node_interface
from tecton_core.query.nodes import exclude_query_node_params
from tecton_spark.query import translate


@attrs.frozen
class SparkExecNode:
    columns: List[str]

    @classmethod
    def from_query_node(cls, query_node: node_interface.QueryNode) -> "SparkExecNode":
        kwargs = attrs.asdict(
            query_node,
            value_serializer=translate.attrs_spark_converter,
            recurse=False,
            filter=exclude_query_node_params,
        )
        kwargs["columns"] = query_node.columns
        del kwargs["dialect"]
        del kwargs["compute_mode"]
        del kwargs["func"]
        del kwargs["node_id"]
        return cls(**kwargs)

    def to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        df = self._to_dataframe(spark)
        if {c.lower() for c in df.columns} != {c.lower() for c in self.columns}:
            pass
            # Because we do not refresh schemas on data sources, we can sometimes get different columns than what we have
            # cached. This is problematic but will require separate solution; don't fail for now
            # raise RuntimeError(f"Returned mismatch of columns: received: {df.columns}, expected: {self.columns}")
        return df

    @abstractmethod
    def _to_dataframe(self, spark: pyspark.sql.SparkSession) -> pyspark.sql.DataFrame:
        raise NotImplementedError



# from tecton import types as sdk_types
import pytest

from tecton import types as sdk_types
from tecton_core.errors import TectonValidationError


def test_field_with_primitives():
    sdk_types.Field("I32_COL", sdk_types.Int32)
    sdk_types.Field("I64_COL", sdk_types.Int64)
    sdk_types.Field("F32_COL", sdk_types.Float32)
    sdk_types.Field("F64_COL", sdk_types.Float64)
    sdk_types.Field("STR_COL", sdk_types.String)
    sdk_types.Field("BOOL_COL", sdk_types.Bool)
    sdk_types.Field("TS_COL", sdk_types.Timestamp)


def test_field_enforces_instantiating_complex_data_types():
    # Passing the class should fail
    with pytest.raises(
        TectonValidationError,
        match="Expected <class 'tecton.types.Array'> to be instance of Array for field asdf, got class object instead",
    ) as exc_info:
        sdk_types.Field("asdf", sdk_types.Array)
    with pytest.raises(
        TectonValidationError,
        match="Expected <class 'tecton.types.Struct'> to be instance of Struct for field asdf, got class object instead",
    ) as exc_info:
        sdk_types.Field("asdf", sdk_types.Struct)
    with pytest.raises(
        TectonValidationError,
        match="Expected <class 'tecton.types.Map'> to be instance of Map for field asdf, got class object instead",
    ) as exc_info:
        sdk_types.Field("asdf", sdk_types.Map)

    # Instances of each class succeed
    sdk_types.Field("asdf", sdk_types.Array(sdk_types.String))
    sdk_types.Field("asdf", sdk_types.Struct(sdk_types.String))
    sdk_types.Field("asdf", sdk_types.Map(sdk_types.String, sdk_types.String))



"""A utility module for Pydantic usages in Tecton.

For compatibility reasons (see TEC-18028), Tecton cannot yet use Pydantic V2. Tecton will use Pydantic V1 whether the
user has V2 or V1 installed locally (in the case of V2 via using the `pydantic.v1` compat library). Tecton code should
reference pydantic solely through this module, e.g.

from tecton._internals.tecton_pydantic import pydantic_v1

This is enforced via pre-submits.

Additionally, this module contains some Pydantic utilities to bridge the gap between Pydantic V1 and V2, e.g. the
StrictModel.
"""

import datetime


try:
    from pydantic import v1 as pydantic_v1
except ImportError:
    import pydantic as pydantic_v1

import sys
import typing
from typing import Any
from typing import Dict
from typing import Type

import typeguard
from typing_extensions import Final


if sys.version_info >= (3, 8):
    from typing import get_args
    from typing import get_origin
else:
    from typing_extensions import get_args
    from typing_extensions import get_origin


class _BannedAnnotationsMeta(type(pydantic_v1.BaseModel)):
    """Block Union annotations that may lead to unwanted type coercion, e.g. Union[str, int], at class definition time.

    Providing an integer to a field with the annotation 'Union[str, int]' would pass the strict type check, but
    pydantic would coerce the integer to string during model construction.
    """

    def __new__(cls, name: str, bases: tuple, namespace: Dict[str, Any], **kwargs: Any) -> Type[pydantic_v1.BaseModel]:
        annotations = namespace.get("__annotations__", {})
        for field_name, field_type in annotations.items():
            if _annotation_has_banned_union(field_type):
                msg = f"Invalid StrictModel annotation: {field_name}. Unions with Pydantic coercible types are not allowed."
                raise TypeError(msg)
        return super().__new__(cls, name, bases, namespace, **kwargs)


class StrictModel(pydantic_v1.BaseModel, metaclass=_BannedAnnotationsMeta):
    """A "strict" pydantic v1 model that blocks extra fields and most Pydantic v1 type coercion.

    Pydantic V1 will coerce values, e.g. coercing integers to datetimes or strings to integers. This generally an
    undesirable behavior for how we use Pydantic. (Pydantic V2 has built-in strict models.)
    """

    @pydantic_v1.validator("*", pre=True)
    def _validate_strict_type(cls, v, field):
        """Check value types before pydantic type coercion. Fails if types do not match according to Typeguard."""
        assert (
            field.annotation is not None
        ), f"Missing annotation for {field.name} for {cls}. StrictModel requires that all fields have annotations."

        if _field_has_model_type(field):
            # Skip strict type checking on nested models. These nested models will run their own type checking.
            # Using strict type checking on these nested models breaks Pydantic deserialization. (This workaround is
            # not needed with Pydantic V2 strict models.)
            return v

        # Some handling for forward references, this will only work for top-level ForwardRefs, e.g. won't work for
        # Optional["Foo"]. Typeguard V4 and/or Pydantic V2 have better options for forward references, but we can't use
        # those due to Python 3.7 and various compatibility issues.
        field_type = field.type_ if isinstance(field.annotation, typing.ForwardRef) else field.annotation

        try:
            typeguard.check_type(field.name, v, field_type)
        except TypeError:
            msg = f"Invalid type for {field.name}. Expected {field.annotation}. Got {v} ({type(v)})."
            raise TypeError(msg)

        return v

    class Config:
        # Do not allow extra attributes during model initialization, i.e. providing kwargs that were not declared
        # fields.
        extra: Final[str] = "forbid"


class StrictFrozenModel(StrictModel):
    """A strict, frozen (i.e. immutable) Pydantic V1 model. See StrictModel."""

    class Config:
        allow_mutation: Final[bool] = False


def _field_has_model_type(field: pydantic_v1.fields.ModelField) -> bool:
    """Returns true if the field is a pydantic model or contains a model within a Union."""

    def is_model_class(annotation: typing.Any) -> bool:
        # Must check that the annotation is a "type" (e.g. a class) because typing annotation (e.g. Union) are not
        # "types".
        return isinstance(annotation, type) and issubclass(annotation, pydantic_v1.BaseModel)

    if get_origin(field.annotation) is typing.Union:
        return any(is_model_class(arg) for arg in get_args(field.annotation))
    else:
        return is_model_class(field.annotation)


_BANNED_UNION_TYPES = (int, float, str, bool, datetime.datetime, datetime.timedelta)


def _annotation_has_banned_union(annotation: typing.Any) -> bool:
    """Returns true if the field has an unsafe Union - i.e. a Union where Pydantic V1 may attempt a coercion."""
    args = get_args(annotation)

    # For Unions, check if any of the arguments (i.e. 'Union[arg1, arg2, ...]' are banned types. Optional is allowed.
    if get_origin(annotation) is typing.Union:
        is_optional = len(args) == 2 and type(None) in args
        contained_banned_type = any(isinstance(arg, type) and arg in _BANNED_UNION_TYPES for arg in args)

        if contained_banned_type and not is_optional:
            return True

    # Recursively check nested types, e.g. Dict[str, Union[int, float]].
    return any(_annotation_has_banned_union(arg) for arg in args)



from dataclasses import dataclass
from datetime import datetime
from datetime import timedelta
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import pytest

from tecton._internals import tecton_pydantic
from tecton._internals.tecton_pydantic import pydantic_v1


@dataclass
class TestTectonPydantic__BasicTypeTestCase:
    annotation: type
    valid_values: Any
    invalid_values: Any


TYPE_TESTS = [
    TestTectonPydantic__BasicTypeTestCase(int, valid_values=[1], invalid_values=[None, 1.0, "1"]),
    TestTectonPydantic__BasicTypeTestCase(Optional[int], valid_values=[1, None], invalid_values=[1.0, "1"]),
    TestTectonPydantic__BasicTypeTestCase(bool, valid_values=[True, False], invalid_values=[None, 1, "True"]),
    # Typeguard, which we use for typechecking Pydantic V1, permits integer as floats.
    TestTectonPydantic__BasicTypeTestCase(float, valid_values=[1.0, 1], invalid_values=[None, "1.0"]),
    TestTectonPydantic__BasicTypeTestCase(str, valid_values=[""], invalid_values=[None, 1]),
    TestTectonPydantic__BasicTypeTestCase(
        datetime, valid_values=[datetime(2022, 1, 1)], invalid_values=[1, "2020-07-10 15:00:00"]
    ),
    TestTectonPydantic__BasicTypeTestCase(timedelta, valid_values=[timedelta(seconds=1)], invalid_values=[1]),
    TestTectonPydantic__BasicTypeTestCase(Any, valid_values=[1, None, ""], invalid_values=[]),
    TestTectonPydantic__BasicTypeTestCase(List[int], valid_values=[[], [1, 2]], invalid_values=[[1, ""]]),
    TestTectonPydantic__BasicTypeTestCase(Tuple[int, str], valid_values=[(1, "foo")], invalid_values=[(1, 2)]),
]


# This test checks a couple of things:
# 1) Some of the exact typechecking behavior that we get from typeguard.
# 2) That the type checking is wired up correct with pydantic, i.e. it's run before typeguard type coercion.
@pytest.mark.parametrize("test_case", TYPE_TESTS)
def test_type_checking(test_case: TestTectonPydantic__BasicTypeTestCase):
    class TestModel(tecton_pydantic.StrictModel):
        field: test_case.annotation

    for valid_value in test_case.valid_values:
        TestModel(field=valid_value)

    for invalid_value in test_case.invalid_values:
        with pytest.raises(pydantic_v1.ValidationError):
            TestModel(field=invalid_value)


@pytest.mark.parametrize("base_type", [tecton_pydantic.StrictModel, tecton_pydantic.StrictFrozenModel])
def test_extra_fields_disallowed(base_type):
    class TestModel(base_type):
        field: int

    with pytest.raises(pydantic_v1.ValidationError):
        TestModel(field=1, extra_field=2)


def test_frozen_model():
    class TestModel(tecton_pydantic.StrictFrozenModel):
        field: int

    instance = TestModel(field=1)

    with pytest.raises(TypeError):
        instance.field = 2


def test_banned_union():
    with pytest.raises(TypeError, match=r"Unions with Pydantic coercible types are not allowed"):

        class TestModel(tecton_pydantic.StrictModel):
            field: Union[int, float]


def test_banned_nested_union():
    with pytest.raises(TypeError, match=r"Unions with Pydantic coercible types are not allowed"):

        class TestModel(tecton_pydantic.StrictModel):
            field: Dict[str, Union[datetime, timedelta]]


class TestForwardDeclaredModel(tecton_pydantic.StrictModel):
    field: "TestForwardDeclaredSubModel"


class TestForwardDeclaredSubModel(tecton_pydantic.StrictModel):
    field: int


TestForwardDeclaredModel.update_forward_refs()


def test_forward_declaration():
    TestForwardDeclaredModel(field=TestForwardDeclaredSubModel(field=1))


def test_nested_model_parsing():
    class SubModel(tecton_pydantic.StrictModel):
        field: int

    class Model(tecton_pydantic.StrictModel):
        submodel: SubModel
        optional_submodel: Optional[SubModel]

    Model(**{"submodel": {"field": 1}, "optional_submodel": {"field": 2}})  # noqa: PIE804



import base64
import contextlib
import logging
import os
from datetime import datetime
from typing import Optional
from urllib.parse import urlparse

import pyarrow.fs
import ray

from tecton_core import conf
from tecton_core.feature_definition_wrapper import FeatureDefinitionWrapper
from tecton_core.query.dialect import Dialect
from tecton_core.query.node_interface import NodeRef
from tecton_core.query.query_tree_compute import SQLCompute
from tecton_core.query.query_tree_executor import QueryTreeExecutor
from tecton_core.query.snowflake.compute import SnowflakeCompute
from tecton_core.secrets import SecretResolver
from tecton_core.snowflake_context import SnowflakeContext
from tecton_materialization.common.job_metadata import JobMetadataClient
from tecton_materialization.common.task_params import feature_definition_from_task_params
from tecton_materialization.ray.batch_materialization import run_batch_materialization
from tecton_materialization.ray.delta import DeltaWriter
from tecton_materialization.ray.feature_export import get_feature_export_qt
from tecton_materialization.ray.feature_export import get_feature_export_store_params
from tecton_materialization.ray.ingest_materialization import ingest_pushed_df
from tecton_materialization.ray.job_status import JobStatusClient
from tecton_materialization.ray.job_status import MonitoringContextProvider
from tecton_materialization.ray.materialization_utils import get_delta_writer
from tecton_materialization.ray.materialization_utils import run_online_store_copier
from tecton_materialization.secrets import MDSSecretResolver
from tecton_proto.materialization.job_metadata__client_pb2 import TectonManagedStage
from tecton_proto.materialization.params__client_pb2 import MaterializationTaskParams
from tecton_proto.materialization.params__client_pb2 import SecretMaterializationTaskParams
from tecton_proto.materialization.params__client_pb2 import SecretServiceParams
from tecton_proto.offlinestore.delta import metadata__client_pb2 as metadata_pb2
from tecton_proto.online_store_writer.copier__client_pb2 import DeletionRequest
from tecton_proto.online_store_writer.copier__client_pb2 import OnlineStoreCopierRequest


logger = logging.getLogger(__name__)

FEATURE_EXPORT_TASK_TYPE = "feature_export"

_DIALECT_TO_STAGE_TYPE = {
    Dialect.PANDAS: TectonManagedStage.PYTHON,
    Dialect.DUCKDB: TectonManagedStage.PYTHON,
    Dialect.SNOWFLAKE: TectonManagedStage.SNOWFLAKE,
}

_DIALECT_TO_UI_STRING = {
    Dialect.PANDAS: "Python",
    Dialect.DUCKDB: "Python",
    Dialect.SNOWFLAKE: "Snowflake",
}


def _get_compute(dialect: Dialect, qt: NodeRef, secret_resolver: Optional[SecretResolver]) -> SQLCompute:
    if dialect == Dialect.SNOWFLAKE:
        if SnowflakeContext.is_initialized():
            return SnowflakeCompute.for_connection(SnowflakeContext.get_instance().get_connection())
        else:
            return SnowflakeCompute.for_query_tree(qt, secret_resolver)
    return SQLCompute.for_dialect(dialect)


def _delete_from_online_store(
    materialization_task_params: MaterializationTaskParams, job_status_client: JobStatusClient
) -> None:
    online_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.ONLINE_STORE,
        "Unload features to online store",
    )
    with online_stage_monitor() as progress_callback:
        if materialization_task_params.deletion_task_info.deletion_parameters.HasField("online_join_keys_path"):
            deletion_request = DeletionRequest(
                online_join_keys_path=materialization_task_params.deletion_task_info.deletion_parameters.online_join_keys_path,
            )
        else:
            deletion_request = DeletionRequest(
                online_join_keys_full_path=materialization_task_params.deletion_task_info.deletion_parameters.online_join_keys_full_path,
            )
        request = OnlineStoreCopierRequest(
            online_store_writer_configuration=materialization_task_params.online_store_writer_config,
            feature_view=materialization_task_params.feature_view,
            deletion_request=deletion_request,
        )
        run_online_store_copier(request)
    progress_callback(1.0)


def _delete_from_offline_store(params: MaterializationTaskParams, job_status_client: JobStatusClient):
    offline_uri = params.deletion_task_info.deletion_parameters.offline_join_keys_path
    fs, path = pyarrow.fs.FileSystem.from_uri(offline_uri)
    keys_table = pyarrow.dataset.dataset(source=path, filesystem=fs).to_table()
    offline_stage_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE, "Delete keys from offline store"
    )
    with offline_stage_monitor():
        delta_writer = get_delta_writer(params)
        delta_writer.delete_keys(keys_table)
        delta_writer.commit()


def _feature_export(
    fd: FeatureDefinitionWrapper,
    task_params: MaterializationTaskParams,
    job_status_client: JobStatusClient,
    executor: QueryTreeExecutor,
):
    export_params = task_params.feature_export_info.feature_export_parameters
    start_time = export_params.feature_start_time.ToDatetime()
    end_time = export_params.feature_end_time.ToDatetime()
    table_uri = export_params.export_store_path
    store_params = get_feature_export_store_params(fd)
    delta_writer = get_delta_writer(task_params, store_params_override=store_params, table_uri_override=table_uri)
    delta_write_monitor = job_status_client.create_stage_monitor(
        TectonManagedStage.StageType.OFFLINE_STORE,
        "Write full features to offline store.",
    )

    _run_export(fd, start_time, end_time, executor, delta_writer, delta_write_monitor)


def _run_export(
    fd: FeatureDefinitionWrapper,
    start_time: datetime,
    end_time: datetime,
    qt_executor: QueryTreeExecutor,
    delta_writer: DeltaWriter,
    write_monitor: MonitoringContextProvider,
):
    qt, interval = get_feature_export_qt(fd, start_time, end_time)
    feature_data = qt_executor.exec_qt(qt).result_table

    with write_monitor():
        # TODO (TEC-18865): add support for is_overwrite flag for feature_export jobs
        transaction_metadata = metadata_pb2.TectonDeltaMetadata()
        transaction_metadata.feature_start_time.FromDatetime(interval.start)

        @delta_writer.transaction(transaction_metadata)
        def txn():
            delta_writer.delete_time_range(interval)
            delta_writer.write(feature_data)

        txn()


@contextlib.contextmanager
def _ray():
    print(f"Initializing Ray from classpath: {os.environ['CLASSPATH']}")
    ray.init(
        job_config=ray.job_config.JobConfig(code_search_path=os.environ["CLASSPATH"].split(":")),
        include_dashboard=False,
    )
    try:
        yield
    finally:
        ray.shutdown()


def run_materialization(
    materialization_task_params: MaterializationTaskParams,
    secret_materialization_task_params: SecretMaterializationTaskParams,
) -> None:
    logger.info(f"Starting materialization job for task: {materialization_task_params.materialization_task_id}")
    conf.set("DUCKDB_DEBUG", "true")
    conf.set("TECTON_OFFLINE_RETRIEVAL_COMPUTE_MODE", "rift")
    conf.set("TECTON_RUNTIME_MODE", "MATERIALIZATION")
    assert materialization_task_params.feature_view.schemas.HasField("materialization_schema"), "missing schema"
    job_status_client = JobStatusClient(JobMetadataClient.for_params(materialization_task_params))
    job_status_client.set_overall_state(TectonManagedStage.State.RUNNING)
    try:
        with _ray():
            run_ray_job(materialization_task_params, secret_materialization_task_params, job_status_client)
            job_status_client.set_overall_state(TectonManagedStage.State.SUCCESS)
    except Exception:
        job_status_client.set_current_stage_failed(TectonManagedStage.ErrorType.UNEXPECTED_ERROR)
        job_status_client.set_overall_state(TectonManagedStage.State.ERROR)
        raise


def _run_maintenance(materialization_task_params: MaterializationTaskParams, fd: FeatureDefinitionWrapper) -> None:
    delta_writer = get_delta_writer(materialization_task_params)
    maintenance_params = materialization_task_params.delta_maintenance_task_info.delta_maintenance_parameters

    if maintenance_params.execute_sorting:
        delta_writer.run_z_order_optimization()
        delta_writer.commit()
    if maintenance_params.vacuum:
        # vacuuming is not implemented
        pass


def run_ray_job(
    materialization_task_params: MaterializationTaskParams,
    secret_materialization_task_params: SecretMaterializationTaskParams,
    job_status_client: JobStatusClient,
) -> None:
    fd = feature_definition_from_task_params(materialization_task_params)
    secret_resolver = _get_secret_resolver(secret_materialization_task_params.secret_service_params)
    task_type = _get_task_type(materialization_task_params)
    executor = QueryTreeExecutor(monitor=job_status_client, secret_resolver=secret_resolver)

    if task_type == "deletion_task":
        _delete_from_offline_store(materialization_task_params, job_status_client)
        _delete_from_online_store(materialization_task_params, job_status_client)
    elif task_type == "delta_maintenance_task":
        # TODO (TEC-19141): delta-rs maintenance commits are incompatible with delta standalone library
        _run_maintenance(materialization_task_params, fd)
    elif task_type == "ingest_task":
        ingest_pushed_df(materialization_task_params, fd, job_status_client)
    elif task_type == FEATURE_EXPORT_TASK_TYPE:
        _feature_export(fd, materialization_task_params, job_status_client, executor)
    elif task_type == "batch_task":
        run_batch_materialization(materialization_task_params, job_status_client, executor)
    else:
        msg = f"Task type {task_type} is not supported by Ray materialization job"
        raise ValueError(msg)


def _get_task_type(materialization_task_params: MaterializationTaskParams) -> str:
    return materialization_task_params.WhichOneof("task_info")[:-5]  # removesuffix("_info")


def _get_secret_resolver(
    secret_service_params: SecretServiceParams,
) -> Optional[SecretResolver]:
    if not secret_service_params.secrets_api_service_url:
        return None

    assert secret_service_params.secret_access_api_key, "Secret access key is required when using secret service"
    return MDSSecretResolver(
        secret_service_params.secrets_api_service_url,
        secret_service_params.secret_access_api_key,
    )


def main():
    params = MaterializationTaskParams()
    secret_params = SecretMaterializationTaskParams()

    # Extract secret m13n params from env vars. VM-based Rift passes secrets via this env var.
    if "SECRET_MATERIALIZATION_PARAMS" in os.environ:
        logger.info("Found secret materialization params, parsing...")
        secret_params.ParseFromString(base64.standard_b64decode(os.environ["SECRET_MATERIALIZATION_PARAMS"]))

    # Legacy way of passing standard m13n params via env vars. Required for compatibility with Anyscale and GCP.
    if "MATERIALIZATION_TASK_PARAMS" in os.environ:
        logger.info("Found legacy materialization task params in env vars, parsing...")
        params.ParseFromString(base64.standard_b64decode(os.environ["MATERIALIZATION_TASK_PARAMS"]))

        # Backwards compatibility for fetching secret access key from materialization task params, if not already found in the secret env var. Anyscale relies on this. Should be removed once Anyscale is completely gone
        # See https://tecton.atlassian.net/browse/TEC-20193
        if params.secrets_api_service_url and not secret_params.HasField("secret_service_params"):
            secret_service_params = SecretServiceParams()
            secret_service_params.secrets_api_service_url = params.secrets_api_service_url
            secret_service_params.secret_access_api_key = params.secret_access_api_key

            secret_params.secret_service_params = secret_service_params

        run_materialization(params, secret_params)
        return

    # Extract standard m13n params from the object store
    if "MATERIALIZATION_TASK_PARAMS_S3_URL" in os.environ:
        parsed = urlparse(os.environ["MATERIALIZATION_TASK_PARAMS_S3_URL"])
        logger.info(f"Found materialization task params at {parsed}, parsing...")
        import boto3

        bucket = parsed.netloc
        key = parsed.path.lstrip("/")
        s3 = boto3.client("s3")
        resp = s3.get_object(Bucket=bucket, Key=key)
        params.ParseFromString(resp["Body"].read())

        run_materialization(params, secret_params)
        return

    msg = "Materialization params were not provided"
    raise ValueError(msg)


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tecton_proto/data/realtime_compute__client.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n0tecton_proto/data/realtime_compute__client.proto\x12\x11tecton_proto.data\"\x18\n\x16\x43olocatedComputeConfig\"d\n\x1bRemoteFunctionComputeConfig\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12!\n\x0c\x66unction_uri\x18\x03 \x01(\tR\x0b\x66unctionUri\"\xdb\x01\n\x17\x41WSComputeInstanceGroup\x12\x32\n\x15\x61utoscaling_group_arn\x18\x01 \x01(\tR\x13\x61utoscalingGroupArn\x12\x34\n\x16\x61utoscaling_group_name\x18\x02 \x01(\tR\x14\x61utoscalingGroupName\x12\x16\n\x06region\x18\x03 \x01(\tR\x06region\x12\x12\n\x04port\x18\x04 \x01(\x05R\x04port\x12*\n\x11health_check_path\x18\x05 \x01(\tR\x0fhealthCheckPath\"\x98\x01\n\x0e\x41WSTargetGroup\x12\x10\n\x03\x61rn\x18\x01 \x01(\tR\x03\x61rn\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12`\n\x16\x63ompute_instance_group\x18\x03 \x01(\x0b\x32*.tecton_proto.data.AWSComputeInstanceGroupR\x14\x63omputeInstanceGroup\"p\n\x1fGoogleCloudComputeInstanceGroup\x12\x18\n\x07project\x18\x01 \x01(\tR\x07project\x12\x16\n\x06region\x18\x02 \x01(\tR\x06region\x12\x1b\n\ttarget_id\x18\x03 \x01(\tR\x08targetId\"\xd4\x01\n\x19GoogleCloudBackendService\x12\x1b\n\ttarget_id\x18\x01 \x01(\tR\x08targetId\x12\x18\n\x07project\x18\x02 \x01(\tR\x07project\x12\x16\n\x06region\x18\x03 \x01(\tR\x06region\x12h\n\x16\x63ompute_instance_group\x18\x04 \x01(\x0b\x32\x32.tecton_proto.data.GoogleCloudComputeInstanceGroupR\x14\x63omputeInstanceGroup\"\xd3\x01\n\x12LoadBalancerTarget\x12M\n\x10\x61ws_target_group\x18\x01 \x01(\x0b\x32!.tecton_proto.data.AWSTargetGroupH\x00R\x0e\x61wsTargetGroup\x12\x64\n\x16google_backend_service\x18\x02 \x01(\x0b\x32,.tecton_proto.data.GoogleCloudBackendServiceH\x00R\x14googleBackendServiceB\x08\n\x06target\"\xb7\x03\n\x13OnlineComputeConfig\x12X\n\x11\x63olocated_compute\x18\x01 \x01(\x0b\x32).tecton_proto.data.ColocatedComputeConfigH\x00R\x10\x63olocatedCompute\x12W\n\x0eremote_compute\x18\x02 \x01(\x0b\x32..tecton_proto.data.RemoteFunctionComputeConfigH\x00R\rremoteCompute\x12i\n\x1a\x61ws_compute_instance_group\x18\x03 \x01(\x0b\x32*.tecton_proto.data.AWSComputeInstanceGroupH\x00R\x17\x61wsComputeInstanceGroup\x12w\n\x1dgoogle_compute_instance_group\x18\x04 \x01(\x0b\x32\x32.tecton_proto.data.GoogleCloudComputeInstanceGroupH\x00R\x1agoogleComputeInstanceGroupB\t\n\x07\x63omputeB\x13\n\x0f\x63om.tecton.dataP\x01')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'tecton_proto.data.realtime_compute__client_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\017com.tecton.dataP\001'
  _COLOCATEDCOMPUTECONFIG._serialized_start=71
  _COLOCATEDCOMPUTECONFIG._serialized_end=95
  _REMOTEFUNCTIONCOMPUTECONFIG._serialized_start=97
  _REMOTEFUNCTIONCOMPUTECONFIG._serialized_end=197
  _AWSCOMPUTEINSTANCEGROUP._serialized_start=200
  _AWSCOMPUTEINSTANCEGROUP._serialized_end=419
  _AWSTARGETGROUP._serialized_start=422
  _AWSTARGETGROUP._serialized_end=574
  _GOOGLECLOUDCOMPUTEINSTANCEGROUP._serialized_start=576
  _GOOGLECLOUDCOMPUTEINSTANCEGROUP._serialized_end=688
  _GOOGLECLOUDBACKENDSERVICE._serialized_start=691
  _GOOGLECLOUDBACKENDSERVICE._serialized_end=903
  _LOADBALANCERTARGET._serialized_start=906
  _LOADBALANCERTARGET._serialized_end=1117
  _ONLINECOMPUTECONFIG._serialized_start=1120
  _ONLINECOMPUTECONFIG._serialized_end=1559
# @@protoc_insertion_point(module_scope)
